class Plot:

    __plot_log = []

    def __init__(self, data_set):
        """
        This class is an interface to the plotting environments of the different sub tables.

        :param data_set:
        :type data_set: Phosphorpy.data.data.DataSet
        """
        self._data_set = data_set

    def magnitude_hist(self, *args, **kwargs):
        """
        Plots a magnitude histogram

        :param args:
        :param kwargs:
        :return:
        """
        self._data_set.magnitudes.plot.hist(*args, **kwargs)

    def sed(self, *args, **kwargs):
        """
        Plots the SED of the given target.

        :param args:
        :param kwargs:
        :return:
        """
        self._data_set.flux.plot.sed(*args, **kwargs)
        self.__plot_log.append((self._data_set.flux.plot.sed, args, kwargs))

    def color_color(self, *args, **kwargs):
        """
        Plots a color color graph of the targets for the given colors

        .. code-block:: python

            ds.plot.color_color(['BP - G', 'G - RP'])

        :param args:
        :param kwargs:
        :return:
        """
        self._data_set.colors.plot.color_color(*args, **kwargs)
        self.__plot_log.append((self._data_set.colors.plot.color_color, args, kwargs))

    def color_hist(self, *args, **kwargs):
        """
        Plot the distribution of colors for the given colors

        :param args:
        :param kwargs:
        :return:
        """
        self._data_set.colors.plot.color_hist(*args, **kwargs)
        self.__plot_log.append((self._data_set.colors.plot.color_hist, args, kwargs))

    def equatorial_coordinates(self, **kwargs):
        """
        Plot the coordinates of the targets in a equatorial coordinate system.
        The plot will be a mollweide projection of the coordinates

        .. code-block:: python

            ds.plot.equatorial_coordinates()

        :param kwargs:
            Coordinate plot arguments (path, marker and color).
            See :meth:`Phosphorpy.data.sub.plot.coordinates.equatorial` for details.
        :return:
        """
        self._data_set.coordinates.plot.equatorial(**kwargs)
        self.__plot_log.append((self._data_set.coordinates.plot.equatorial, None, None))

    def galactic_coordinates(self, **kwargs):
        """
        Plot the coordinates of the targets in a galactic coordinate system.
        The plot will be a mollweide projection of the coordinates

        :param kwargs:
            Coordinate plot arguments (path, marker and color).
            See :meth:`Phosphorpy.data.sub.plot.coordinates.galactic` for details.
        :return:
        """
        self._data_set.coordinates.plot.galactic(**kwargs)
        self.__plot_log.append((self._data_set.coordinates.plot.galactic, None, None))



import configparser
import sqlite3
import warnings

import numba
import numpy as np
from armapy.svo import get_survey_filter_information
from astropy.table import Table
from pandas import DataFrame

from Phosphorpy.core.structure import Mask
from .colors import Colors
from .flux import FluxTable
from .plots.magnitude import MagnitudePlot
from .table import DataTable
from .tables.magnitude import Magnitude


def power_2_10(x):
    return np.power(10., -x/2.5)


@numba.vectorize
def subtract(a, b):
    """
    Numba implementation of a subtraction

    :param a:
    :param b:
    :return:
    """
    return a-b


def get_prefix(cols, unique_col):
    """
    Extracts a prefix from the column names

    :param cols: The name of the columns
    :type cols: list
    :param unique_col: A unique column
    :type unique_col: str
    :return: The prefix of all columns
    :rtype: str
    """
    for c in cols:
        if unique_col in c:
            return c.split(unique_col)[0]
    return ''


def get_survey_cols(cols, s_cols, prefix):
    """
    Returns the names of the magnitudes in a specific survey

    :param cols:
    :param s_cols:
    :param prefix:
    :return:
    """
    out = []
    prefix_cond = len(prefix) == 0
    for i, c in enumerate(cols):
        for j, s in enumerate(s_cols):
            # if the survey magnitude name is in and the prefix is in
            # of if there is no prefix and the magnitude name is equal to the survey name
            if ((s in c and prefix in c and not prefix_cond) or
                    (prefix_cond and (c == s or 'mag' in c or 'Petro' in c))):
                out.append(c)
                break

    cols_new = []
    for i, c in enumerate(cols):
        if c not in out:
            cols_new.append(c)
    return cols_new, out


def guess_surveys(mag_cols):
        """
        Try to find matches of the magnitude column names with a few
        implement large sky surveys.

        example:
            SDSS has the column u, g, r, i and z and PAN-STARRS g, r, i, z, y
            so both have 5 bands but PAN-STARRS doesn't have u and SDSS
            doesn't have y.

        :return:
        """
        pan_starrs = 0
        two_mass = 0
        sdss = 0
        kids = 0
        viking = 0
        wise = 0
        apass = 0
        gaia = 0
        galex = 0
        ukidss = 0
        skymapper = 0

        # count how often a band could be part of a survey
        for mag in mag_cols:
            # ignore error columns or J2000 coordinate columns
            if 'e_' in mag or 'J2000' in mag or 'deg' in mag or 'ra' in mag or 'dec' in mag:
                continue
            # strip mag postfix
            if 'mag' in mag:
                mag = mag.split('mag')[0]
            elif 'Petro' in mag:
                mag = mag.split('Petro')[0]

            if 'u' in mag:
                kids += 1
                sdss += 1
                skymapper += 1
            elif 'g' in mag or 'r' in mag or 'i' in mag:
                if 'p' in mag:
                    apass += 1
                else:
                    pan_starrs += 1
                    sdss += 1
                    kids += 1
                    skymapper += 1

            elif 'z' in mag:
                pan_starrs += 1
                sdss += 1
                skymapper += 1
            elif 'y' in mag:
                pan_starrs += 1
            elif 'Y' in mag or 'Z' in mag:
                viking += 1
                ukidss += 1
            elif 'J' in mag or 'H' in mag or 'K' in mag:
                viking += 1
                two_mass += 1
                ukidss += 1
            elif 'W' in mag:
                wise += 1
            elif 'G' in mag or 'RP' in mag or 'BP' in mag:
                gaia += 1
            elif 'FUV' in mag or 'NUV' in mag:
                galex += 1
            elif 'B' in mag or 'V' in mag:
                apass += 1
            elif 'v' in mag:
                skymapper += 1
        # TODO: use data from local file
        # decide which survey it is
        if skymapper == 6:
            kids -= 4
            sdss -= 5
            pan_starrs -= 4
            prefix = get_prefix(mag_cols, 'v')
            s_cols = ['u', 'v', 'g', 'r', 'i', 'z']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            return cols

        if pan_starrs == 5 or (pan_starrs > 5 and 'y' in mag_cols):
            kids -= 3
            sdss -= 4

            prefix = get_prefix(mag_cols, 'y')

            s_cols = ['g', 'r', 'i', 'z', 'y']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            return cols

        if sdss == 5 or sdss == 9:
            kids -= 4

            prefix = get_prefix(mag_cols, 'z')

            s_cols = ['u', 'g', 'r', 'i', 'z']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            return cols

        if kids == 4:
            prefix = get_prefix(mag_cols, 'u')

            s_cols = ['u', 'g', 'r', 'i']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            return cols

        if viking == 5 or (viking > 5 and 'Z' in mag_cols):
            prefix = get_prefix(mag_cols, 'Z')

            s_cols = ['Z', 'Y', 'J', 'H', 'K']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            two_mass -= 3
            ukidss -= 4
            return cols

        if ukidss == 3:
            # todo: update after J band is included
            prefix = get_prefix(mag_cols, 'Y')

            s_cols = ['Y', 'H', 'K']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            two_mass -= 2
            return cols

        if two_mass == 3:
            prefix = get_prefix(mag_cols, 'H')
            s_cols = ['J', 'H', 'K']
            mag_cols, cols = get_survey_cols(mag_cols, s_cols, prefix)
            return cols

        if wise == 4:
            cols = []
            for i, c in enumerate(mag_cols):
                if 'W' in c:
                    cols.append(c)
            return cols

        if gaia == 3:
            cols = []
            n_cols = []
            for i, c in enumerate(mag_cols):
                if 'G' in c or 'BP' in c or 'RP' in c:
                    cols.append(c)
                else:
                    n_cols.append(c)
            return cols

        if apass == 5:
            cols = []
            n_cols = []
            for i, c in enumerate(mag_cols):
                if 'gp' in c or 'rp' in c or 'ip' in c or 'B' in c or 'V' in c:
                    cols.append(c)
                else:
                    n_cols.append(c)
            return cols

        if galex == 2:
            cols = []
            n_cols = []
            for i, c in enumerate(mag_cols):
                if 'FUV' in c or 'NUV' in c:
                    cols.append(c)
                else:
                    n_cols.append(c)
            return cols


def mag_cols_only(cols):
    """
    Returns only the columns names of the magnitudes by excluding columns with "e\_" in it.

    :param cols: A list with all column names
    :type cols: list
    :return: A list with the magnitude names only
    :rtype: list
    """
    out = []
    for c in cols:
        if 'e_' not in c:
            out.append(c)
    return out


class SurveyData:
    """
    Class to handle the properties of different surveys, like providing the information of the different
    bands (flux zero point, effective wavelength, etc).
    """
    _survey_cols = None
    _properties = None

    def __init__(self, name=None, mag_cols=None):
        self._survey_cols = {}
        self._properties = {}
        if name is not None:
            mag_cols = mag_cols_only(mag_cols)
            self.__add_survey_properties__(name, mag_cols)

    def __add_survey_properties__(self, survey, cols):
        survey = survey.lower()
        survey_properties = {}
        for c in cols:
            if 'mag' in c:
                c = c.split('mag')[0]
            elif 'Petro' in c:
                c = c.split('Petro')[0]
            survey_properties[c] = get_survey_filter_information(survey, c)
        self._properties[survey] = survey_properties
        self._survey_cols.update({survey: cols})

    def __getitem__(self, item):
        item = item.lower()
        return self._survey_cols[item]

    def all_magnitudes(self):
        """
        Returns all magnitude column names

        :return:
        """
        out = []
        for s in self._survey_cols:
            out.extend(self._survey_cols[s])
        return out

    def get_survey_id(self, survey_name):
        survey_name = survey_name.lower()
        return np.where(
            np.array(list(self._survey_cols.keys())) == survey_name)[0][0]

    def get_corresponding_survey(self, mag_name):
        for s in self._survey_cols:
            if mag_name in self._survey_cols[s]:
                return s

    def get_corresponding_survey_id(self, mag_name):
        return self.get_survey_id(self.get_corresponding_survey(mag_name))

    def get_survey_magnitude(self, survey_name):
        """
        Returns the names of the magnitudes of the different surveys

        :param survey_name: The name of the survey
        :type survey_name: str
        :return: A list with the names of the magnitudes
        :rtype: list
        """
        survey_name = survey_name.lower()
        return self._survey_cols[survey_name]

    def all_error_names(self):
        """
        Returns all error column names
        :return:
        """
        out = []
        for s in self._survey_cols:
            for c in self._survey_cols[s]:
                out.append('e_{}'.format(c))
        return out

    def get_survey_error_names(self, survey_name):
        survey_name = survey_name.lower()
        out = self.get_survey_magnitude(survey_name)
        o = []
        for k in out:
            o.append('e_{}'.format(k))
        return o

    def get_surveys(self):
        """
        Returns the names of the surveys
        :return:
        """
        return list(self._survey_cols.keys())

    def add_survey(self, name, mag_cols):
        """
        Adds a new survey with the magnitude names to the current list.

        :param name: The name of the new survey
        :param name: str
        :param mag_cols: A list with the magnitude names
        :type mag_cols: list
        :return:
        """
        name = name.lower()
        mag_cols = mag_cols_only(mag_cols)
        self._survey_cols[name] = mag_cols
        self.__add_survey_properties__(name, mag_cols)

    def effective_wavelength(self, survey, band):
        """
        Returns the effective wavelength of the filter

        :param survey: The name of the survey
        :type survey: str
        :param band: The filter band of the survey
        :type band: str
        :return: The effective wavelength
        :rtype: float
        """
        survey = survey.lower()
        lambda_eff = self._properties[survey][band]['lambda_eff']
        try:
            return float(lambda_eff)
        except ValueError:
            return lambda_eff

    def get_all_wavelengths(self):
        """
        Returns all effective wavelengths at once.
        The order is the same order as the magnitude columns.

        :return: All effective wavelengths
        :rtype: numpy.ndarray
        """
        out = []
        for s in self._properties:
            survey = self._properties[s]
            for band in survey:
                out.append(self.effective_wavelength(s, band))
        return np.float64(out)

    def get_survey_wavelengths(self, survey_name):
        """
        Returns the effective wavelengths of the bands of the survey sorted by their values.

        :param survey_name: The name of the survey
        :type survey_name: str
        :return: The wavelengths
        :rtype: numpy.ndarray
        """
        survey_name = survey_name.lower()

        out = []
        survey = self._properties[survey_name]
        for band in survey:
            out.append(self.effective_wavelength(survey_name, band))
        return np.array(out)

    def flux_zero(self, survey, band):
        """
        Returns the flux zero point of the filter.
        For 2MASS, VIKING, GALEX and Gaia the Vega zero point is used and for all the other
        ones, the AB zero point

        :param survey: The name of the survey
        :type survey: str
        :param band: The filter band of the survey
        :type band: str
        :return: The flux zero point in ergs
        :rtype: float
        """
        band = band.split('mag')[0]
        props = self._properties[survey][band]
        # check if in the SVO data the vega system is set as the used system
        if 'vega' in props.keys() and bool(props['vega']):
            return float(props['Vega_ergs'])
        return float(props['AB_ergs'])

    def is_vega(self, survey, band):
        """
        Returns true, if the photometric system for this survey and band is Vega, else False
        :param survey: The name of the survey
        :type survey: str
        :param band: The name of the band
        :type band: str
        :return: True, if the photometric system is vega, else False
        :rtype: bool
        """
        survey = survey.lower()
        return bool(self._properties[survey][band])

    def _to_config(self, path):
        conf = configparser.ConfigParser()
        for survey in self.get_surveys():
            conf.add_section(survey)
            conf.set(survey, 'magnitudes', str(self.get_survey_magnitude(survey)))
            for props in self._properties[survey]:
                proper = self._properties[survey][props]
                for p in proper:
                    conf.set(survey, '{}_{}'.format(props, p), str(proper[p]))
        with open(path, 'w') as f:
            conf.write(f)

    def to_dataframe(self):
        surveys = self.get_surveys()
        data = {'survey': [],
                'band': []}
        for s in surveys:
            bands = self._properties[s]
            for b in bands:
                data['survey'].append(s)
                data['band'].append(b)
                props = bands[b]
                for p in list(props):
                    if p in data.keys():
                        data[p].append(props[p])
                    else:
                        data[p] = [props[p]]
        return DataFrame(data)

    def add_to_meta(self, meta, name):
        bands = self._properties[name]
        meta['name'] = name
        for b in bands:
            props = bands[b]
            for p in list(props):
                meta[f'{b}_{p}'] = props[p]

    def _to_sql(self, path):
        """
        Writes the data to a sql-database

        :param path: Path to the sql-database
        :type path: str
        :return:
        """
        con = sqlite3.Connection(path)
        data = self.to_dataframe()
        data.to_sql('survey_properties', con)

    def write(self, path, data_format='config'):
        """
        Writes the survey data to a config-file

        :param path: Path to the config file
        :type path: str
        :param data_format: Not used
        :return:
        """
        data_format = data_format.lower()
        if data_format == 'config' or data_format == 'init':
            self._to_config(path)
        elif data_format == 'sql':
            self._to_sql(path)
        else:
            raise ValueError(f'Unsupported data format: {data_format}')

    @staticmethod
    def read(path):
        """
        Reads survey data from a config file

        :param path: Path to the config file
        :type path: str
        :return: The survey data from the config file
        :rtype: SurveyData
        """
        conf = configparser.ConfigParser()
        conf.read(path)

        sd = SurveyData()
        for s in conf.sections():
            mag_cols = conf[s]['magnitudes']
            mag_cols = mag_cols.split('[\'')[-1].split('\']')[0]
            mag_cols = mag_cols.split('\', \'')
            props = {}
            for c in conf[s]:
                if 'magnitudes' not in c:
                    band = c.split('_')[0]
                    key = c.split(band+'_')[-1]
                    if band not in props:
                        props[band] = {}
                    props[band][key] = conf[s][c]

            sd._survey_cols[s] = mag_cols
            sd._properties[s] = props

        return sd

    def __str__(self):
        s = ''
        if self._survey_cols is None:
            return 'No information'
        for k in self._survey_cols:
            cols = self._survey_cols[k]
            s += '{}\n'.format(k)
            for c in cols:
                s += '\t{}\n'.format(c)
        return s


def __check_input__(data, names):
    """
    Checks the input data if they have a proper format and covert them (if necessary) to a
    pandas DataFrame

    :param data: The input data
    :param names: The names of the columns
    :return: A DataFrame with the input data
    :rtype: pandas.DataFrame
    """
    if type(data) != DataFrame:
        if type(data) == Table:
            data = data.to_pandas()
        elif type(data) == np.ndarray:
            if len(data.shape) == 1:
                data = DataFrame.from_records(data)
            else:
                if names is None:
                    raise ValueError('If the input data are in numpy.ndarray, names are needed!')
                elif len(names) != data.shape[1]:
                    raise AttributeError('Number of columns does not the number of names!')
                else:
                    data = DataFrame(data=data, columns=names)
    return data


class SurveyMag(DataTable):
    pass

    def __init__(self, data, names=None, survey='', mask=None):
        DataTable.__init__(self, mask=mask)


class MagnitudeTable(DataTable):
    """
    Class to handle any interactions with magnitudes of multiple surveys
    """
    _mag_cols = None
    _err_cols = None
    _survey = None

    def __init__(self, data=None, names=None, survey='', mask=None):
        DataTable.__init__(self, mask=mask)
        self._mag_cols = []
        self._err_cols = []
        self._plot = MagnitudePlot(self)
        data = __check_input__(data, names)

        try:
            cols = guess_surveys(data.columns)
            if cols is not None and survey != '':
                self._survey = SurveyData(survey, cols)
                self._set_cols(cols)
                data = data[cols]
                if type(data) != Magnitude:
                    data = Magnitude(data, survey, mask=mask)
                self._data = [data]
        except ValueError:
            warnings.warn('No magnitude columns!')
        except AttributeError:
            pass

    def __str__(self):
        string = 'Available magnitudes\n'
        for d in self._data:
            string += str(d)
        return string

    def __getitem__(self, item):
        if type(item) == str:
            return self.get_survey_data(item)
        raise ValueError('No indices supported')

    def _set_cols(self, cols):
        mag = []
        err = []
        for c in cols:
            if 'e_' in c:
                err.append(c)
            else:
                mag.append(c)
        self._mag_cols = mag
        self._err_cols = err

    def apply_extinction_correction_to_survey(self, correction, survey):
        """
        Applies the extinction correction to one specific survey

        :param correction:
            Table with the extinction correction.
            The number of rows must be the same as the number of detections in the survey and the columns
            must have the same names as the magnitude columns in the survey
        :type correction: Table, DataFrame
        :param survey: The name of the survey
        :type survey: str
        :return:
        """
        for d in self.data:
            if d.survey_name == survey:
                d.apply_extinction_correction(correction)
                break

    def apply_extinction_correction(self, correction):
        """
        Applies the extinction correction to the data

        :param correction: A table with the correction values
        :type correction: astropy.table.Table, pandas.DataFrame
        :return:
        """
        for d in self.data:
            d.apply_extinction_correction(correction)

    def get_names(self):
        """
        Returns the names of the magnitude columns

        :return: The name of magnitude columns
        :rtype: list
        """
        names = [d.survey_name for d in self.data]
        return names

    def get_flux(self):
        """
        Returns the flux of the columns
        :return:
        """
        if self.data is None:
            raise AttributeError('No photometric data available!\nDownload data first!')
        flux = []
        for d in self.data:
            flux.append(d.get_flux())

        # return a merged table out of the fluxes and the flux errors
        return FluxTable(flux,
                         survey_head=self._survey, mask=self._mask)

    def get_colors(self, cols=None):
        """
        Computes all possible colors. The first magnitude will be the bluer and the second one will be the redder one,
        except magnitude columns are given, then this order is not fixed.

        .. math:

            C_{i, j} = mag_i - mag_j \\
            \\lambda(band_i) < \\lambda(band_j)

        :param cols:
        :return:
        """
        if cols is None:
            cols = []
            for s in self._survey.get_surveys():
                cols.append((s, self._survey.get_survey_magnitude(s)))
        else:
            try:
                cols[0][0][0]
            except TypeError:
                cols = [cols]

        colors = []
        for d in self._data:
            colors.append(d.get_colors())

        return Colors(colors, mask=self._mask)

    def create_mask(self, overwrite=False):
        """
        Creates a new mask object for all data

        :param overwrite:
            True, if the old mask should be replaced, else False. Default is false.
            If a mask is set already and overwrite False, a RuntimeError will raise.
        :type overwrite: bool
        :return:
        """
        if self._mask is not None and not overwrite:
            raise RuntimeError('A mask is set and overwrite is false.')

        if self._data is None:
            raise ValueError("No data were set.")

        indices = []
        for d in self._data:
            indices.extend(d.data.index.values)
        indices = np.unique(indices)

        self._mask = Mask(len(indices), ids=indices)

    def has_full_photometry(self, survey, previous=True):
        """
        Creates a mask of sources, which have photometry values in every band of the survey.

        :param survey: The name of the survey
        :type survey: str
        :param previous: True if the previous mask should be used True, else False. Default is True.
        :type previous: bool
        :return:
        """
        mask = None
        mag_cols = self.survey.get_survey_magnitude(survey)
        d = self._data[self.survey.get_survey_id(survey)]

        for c in mag_cols:

            k = d[c] == d[c]
            if mask is None:
                mask = k
            else:
                mask = mask & k
        try:
            self.mask.add_mask(mask, 'Mask sources with full photometry in {}'.format(survey), combine=previous)
        except AttributeError:
            warnings.warn('No mask was set. Call \'create_mask()\' to create a mask object.')

    def set_limit(self, band, survey=None, minimum=99, maximum=-99, previous=True):
        """
        Sets a magnitude limit to the magnitude columns with the name band.

        .. code-block:: python

            # select all sources with a G magnitude between 16 and 18
            # (notice that minimum and maximum are magnitudes, therefore the minimum value is larger than the maximum)
            ds.magnitudes.set_limit('G', minimum=18, maximum=16)


        :param band: The name of the band/magnitude column
        :type band: str
        :param survey: 
            The name of the specific survey or None. If None all surveys with such a magnitude name will
            used for the limiting
        :type survey: str, None
        :param minimum: The minimal magnitude value. Default is 99, which means that no cut will be done.
        :type minimum: float
        :param maximum: The maximal magnitude value. Default is -99, which means that no cut will be done.
        :type maximum: float
        :param previous: True if the previous mask should be used True, else False. Default is True.
        :type previous: bool
        :return:
        """
        for d in self._data:
            if survey is None or survey == d.survey_name:
                try:
                    d.set_limit(band, minimum=minimum, maximum=maximum, previous=previous)
                except ValueError:
                    pass

    def get_survey_data(self, name):
        """
        Returns the survey data of the survey with the given name.

        :param name: The name of the survey
        :type name: str
        :return: The survey data
        :rtype: SurveyData
        """
        for d in self._data:
            if name == d.survey_name or name.lower() == d.survey_name:
                return d

    def get_columns(self, cols):
        """
        Creates a (sub)set of the data with the given columns

        :param cols: The name of the magnitude columns
        :type cols: list, str
        :return: DataFrame with the magnitude information or None if no columns were vound
        :rtype: pd.DataFrame
        """
        if type(cols) == str:
            cols = [cols]
        elif type(cols) != list:
            raise AttributeError('"cols" must be a string or a list.')

        out = None
        for d in self.data:
            rs = d.get_columns(cols)
            if len(rs.columns) > 0:
                if out is None:
                    out = rs
                else:
                    out = out.join(rs, rsuffix=f'_{d.survey_name}')
        return out

    @property
    def mag_names(self):
        return self._mag_cols

    @property
    def err_names(self):
        return self._err_cols

    @property
    def survey(self):
        return self._survey

    @survey.setter
    def survey(self, value):
        self._survey = value

    @property
    def head(self):
        return self._survey

    def add_survey_mags(self, mags, survey):
        """
        Adds a new set of magnitudes from another survey. If already magnitudes with the same name are
        in the current data, all the new columns will be renamed to the magnitude name plus the
        survey name.

        :param mags: The new magnitudes
        :param survey: The name of the new survey
        :type survey: str
        :return:
        """
        # cols = guess_surveys(mags.columns)
        cols = []
        for c in mags.columns:
            cl = c.lower()
            # todo: magnitude columns from config file
            if 'ra' in cl or 'de' in cl or 'plx' in cl:
                continue
            cols.append(c)
        mags = Magnitude(mags, cols, survey, mask=self.mask)
        mags.select_columns(cols)

        if len(mags) == 0:
            warnings.warn(f'No data found in {survey} for the given coordinates.\n{survey} will not be listed!')
            return

        survey = survey.lower()

        if self._data is not None:
            try:
                self._data.append(mags)
                self._survey.add_survey(survey, cols)
            # if the magnitudes include magnitudes with the same name already
            except ValueError:
                # create a map to rename the all columns of the new survey with the survey name
                # as postfix
                name_map = {}
                for c in cols:
                    name_map[c] = '{}_{}'.format(c, survey)

                mags.rename(name_map, axis='columns')

                # join the data
                self._data.append(mags)
                self._survey.add_survey(survey, list(name_map.values()))
        # if no magnitudes are set
        else:
            self._data = [mags]

            self._survey = SurveyData(survey, mags.columns)
        mags.set_survey_data(self._survey)

    def write(self, path, data_format='parquet', **kwargs):
        out = []
        for d in self.data:
            out.append(
                d.write(path, data_format=data_format)
            )
        return out



import inspect

import numpy as np
from astropy import units as u
from astropy.units.quantity import Quantity
from pandas import DataFrame
from scipy.constants import h, k, c
from scipy.optimize import curve_fit

from Phosphorpy.data.sub.plots.flux import FluxPlot
from Phosphorpy.data.sub.tables.flux import Flux
from .table import DataTable


def get_column_names(length):
    # create columns names 'a', 'b', 'c', ...
    cols = []
    for i in range(97, 98 + length):
        cols.append(chr(i))
    return cols


def _apply_wavelength_limits(wavelengths, d, err, lower_limit, upper_limit):
    """
    Applies wavelength limits to the data

    :param wavelengths: The wavelengths of the d columns
    :type wavelengths: Union
    :param d:
        The data, rows are the different samples and columns are the data-points
        for every sample
    :type d: Union
    :param err:
        The errors of the data. The shape must be the same as d's shape.
    :type err: Union
    :param lower_limit: The lower wavelength limit
    :type lower_limit: float, Quantity
    :param upper_limit: The upper wavelength limit
    :type upper_limit: float, Quantity
    :return: The wavelengths, data and errors in the wavelength limit
    :rtype: Union, Union, Union
    """
    if lower_limit is not None:
        if type(lower_limit) != Quantity:
            lower_limit = lower_limit*u.angstrom

        wavelength_mask = (wavelengths*u.angstrom > lower_limit)
        d = d[:, wavelength_mask]
        err = err[:, wavelength_mask]
        wavelengths = wavelengths[wavelength_mask]

    if upper_limit is not None:
        if type(upper_limit) != Quantity:
            upper_limit = upper_limit*u.angstrom
        wavelength_mask = (wavelengths*u.angstrom < upper_limit)
        d = d[:, wavelength_mask]
        err = err[:, wavelength_mask]
        wavelengths = wavelengths[wavelength_mask]
    return wavelengths, d, err


def _error_weighted_fitting(wavelengths, d, err, degree):
    """
    Fit all data with a polynomial. if the data or the errors contain NaN values,
    then they will be excluded from the fitting process.

    If the fit fails, all values are filled with -99999.

    :param wavelengths: The wavelengths of the d columns
    :type wavelengths: Union
    :param d:
        The data, rows are the different samples and columns are the data-points
        for every sample
    :type d: Union
    :param err:
        The errors of the data. The shape must be the same as d's shape.
    :type err: Union
    :param degree: The degree of the polynomial
    :type degree: int
    :return: The coefficients from the fit's.
    :rtype: list
    """
    fits = []
    for row, err_row in zip(d, err):
        try:
            m = (row > -9999) & (err_row > -999)
            fit = np.polyfit(wavelengths[m], row[m], degree, w=1/err_row[m])
            fits.append(fit)
        except np.linalg.LinAlgError:
            fits.append((degree+1)*[-99999])
        except TypeError:
            fits.append((degree+1)*[-99999])
    return fits


def _fitting(wavelengths, d, degree):
    """
    Fit all data with a polynomial. if the data contain NaN values, then they will
    be excluded from the fitting process.

    If the fit fails, all values are filled with -99999.

    :param wavelengths: The wavelengths of the d columns
    :type wavelengths: Union
    :param d:
        The data, rows are the different samples and columns are the data-points
        for every sample
    :type d: Union
    :param degree: The degree of the polynomial
    :type degree: int
    :return: The coefficients from the fit's.
    :rtype: list
    """
    fits = []
    for row in d:
        try:
            m = (row > -9999)
            fit = np.polyfit(wavelengths, row[m], degree)
            fits.append(fit)
        except np.linalg.LinAlgError:
            fits.append((degree+1)*[-99999])
        except TypeError:
            fits.append((degree+1)*[-99999])
    return fits


def _normalize(d, err, norm):
    """
    If norm is true, normalize the data by their maximum value otherwise do
    nothing.

    :param d:
        The data, rows are the different samples and columns are the data-points
        for every sample
    :type d: Union
    :param err:
        The errors of the data. The shape must be the same as d's shape.
    :type err: Union
    :param norm:
        True if the data should be normalized by their maximum value else False.
        Default is False
    :return:
        If norm is True, the normalized data, if norm is False, the input data
    :rtype: Union, Union
    """
    if norm:
        maxi = np.nanmax(d, axis=1)
        maxi = np.transpose(np.tile(maxi, (d.shape[1], 1)))
        d = d/maxi
        err = err/maxi
    return d, err


def _rescale(wavelengths, d, err, x_log, y_log):
    """
    Rescales the wavelengths and/or the flux to a logarithmic scale.

    :param wavelengths: The wavelengths of the d columns
    :type wavelengths: Union
    :param d:
        The data, rows are the different samples and columns are the data-points
        for every sample
    :type d: Union
    :param err:
        The errors of the data. The shape must be the same as d's shape.
    :param x_log: If True the wavelengths will become logarithmic.
    :type x_log: bool
    :param y_log: If True, the flux and its errors will become logarithmic.
    :type y_log: bool
    :return:
        If x_log and/or y_log are true, the logarithmic values, else
        the input values
    :rtype: Union, Union, Union
    """
    if x_log:
        wavelengths = np.log10(wavelengths)

    if y_log:
        err = np.abs(1 / d) * err/np.log(10)
        d = np.log10(d)

    return wavelengths, d, err


def _blackbody(lam, temperature, a):
    """ Blackbody as a function of wavelength (um) and temperature (K).

    returns units of erg/s/cm^2/cm/Steradian
    """
    o = 2*h*c**2 / (lam**5 * (np.exp(h*c / (lam*k*temperature)) - 1))
    o /= np.max(o)
    o *= a
    return o


class FluxTable(DataTable):
    """
    FluxTable is the interface to interact with the fluxes of the sources.
    """

    _fits = None
    _data = None
    _survey = None

    def __init__(self, data=None, survey_head=None, mask=None):
        """

        :param data:
        :param survey_head: Survey information
        :type survey_head: SearchEngine.data.sub.magnitudes.SurveyData
        """
        DataTable.__init__(self, mask=mask)
        if data is None:
            data = []
        self._data = data
        self._survey = survey_head
        self._plot = FluxPlot(self)

    def add_fluxes(self, data, survey_name):
        fl = Flux(data, survey=survey_name)
        self.data.append(fl)

    def get_all_fluxes(self):
        """
        Returns all fluxes in a numpy.ndarray with the rows as the sources and the columns as the different bands.
        The order of the bands is the same as the input order.

        :return: The fluxes of all sources
        :rtype: numpy.ndarray
        """
        flux = None
        for d in self.data:
            if flux is None:
                flux = d.get_fluxes()
            else:
                flux = flux.merge(d.get_fluxes(), how='outer', right_index=True, left_index=True)
        return flux.values

    def get_all_errors(self):
        """
        Return all flux uncertainties in a numpy.ndarray with the rows as the sources and the columns as the
        different bands.
        The order of the bands is the same as the input order.

        :return: The flux uncertainties of all sources
        :rtype: numpy.ndarray
        """
        errs = None
        for d in self.data:
            if errs is None:
                errs = d.get_errors()
            else:
                errs = errs.merge(d.get_errors(), how='outer', right_index=True, left_index=True)
        return errs.values

    def get_index(self):
        index = None
        for d in self.data:
            if index is None:
                index = d.get_index()
            else:
                index = index.merge(d.get_index(), how='outer', right_index=True, left_index=True)
        return index.index.values

    def _get_sed_data(self):
        """
        Returns the data of the SED's in structured way.
        First wavelength, then flux and as the last component the errors

        :return: wavelength, flux, flux error
        :rtype: Union, Union, Union
        """
        d = self.get_all_fluxes()
        err = self.get_all_errors()

        wavelengths = self._survey.get_all_wavelengths()

        return wavelengths, d, err

    def get_sed(self, index):
        """
        Returns the SED of the source with the specific index

        :param index: The index of the source
        :type index: int
        :return: The flux and the uncertainties of the specific source
        :rtype: numpy.ndarray, numpy.ndarray
        """
        fluxes = []
        errs = []
        for d in self.data:
            survey_flux, survey_err = d.get_sed(index)
            fluxes.extend(survey_flux)
            errs.extend(survey_err)
        return np.array(fluxes), np.array(errs)

    def fit_polynomial(self, degree, error_weighted=True,
                       x_log=False, y_log=False,
                       norm=False, lower_limit=None, upper_limit=None):
        """
        Fit's all flux SED's with polynomial

        If the fit fails, the values of the coefficients are set to -99999.

        :param degree: The degree of the polynomial
        :type degree: int
        :param error_weighted:
            True if the error should be used as weights, else False.
            Default is True.
        :type error_weighted: bool
        :param x_log:
            True if the wavelengths should be logarithmic, else False.
            Default is False.
        :type x_log: bool
        :param y_log:
            The if the flux should be logarithmic, else False.
            Default is False.
        :type y_log: bool
        :param norm:
            The if the SED should be normalized by its maximum value, else False.
            Default is False.
        :type norm: bool
        :param lower_limit:
            The lower wavelengths limit or None if no lower limit should be applied.
            Default is None.
        :type lower_limit: None, float, Quantity
        :param upper_limit:
            The upper wavelengths limit or None if no upper limit should be applied.
            Default is None.
        :type upper_limit: None, float, Quantity
        :return: The results of the fit's
        :rtype: DataFrame
        """
        wavelengths, d, err = self._get_sed_data()

        # apply wavelength cuts
        wavelengths, d, err = _apply_wavelength_limits(wavelengths, d, err,
                                                       lower_limit, upper_limit)

        wavelengths, d, err = _rescale(wavelengths, d, err, x_log, y_log)

        d, err = _normalize(d, err, norm)

        if error_weighted:
            fits = _error_weighted_fitting(wavelengths, d, err, degree)
        else:
            fits = _fitting(wavelengths, d, degree)

        self._fits = DataFrame(data=np.array(fits),
                               columns=get_column_names(degree),
                               index=self.get_index())

        return self._fits

    def func_fit(self, func, error_weighted=True,
                 x_log=False, y_log=False,
                 norm=False, lower_limit=None, upper_limit=None,
                 **curve_fit_parameters):
        """
        Fit's the wavelength and fluxes with the given function and stores the results in fits

        :param func: The fitting function in the style of scipy.optimize.curve_fit
        :param error_weighted:
            True if the error should be used as weights, else False.
            Default is True.
        :type error_weighted: bool
        :param x_log:
            True if the wavelengths should be logarithmic, else False.
            Default is False.
        :type x_log: bool
        :param y_log:
            The if the flux should be logarithmic, else False.
            Default is False.
        :type y_log: bool
        :param norm:
            The if the SED should be normalized by its maximum value, else False.
            Default is False.
        :type norm: bool
        :param lower_limit:
            The lower wavelengths limit or None if no lower limit should be applied.
            Default is None.
        :type lower_limit: None, float, Quantity
        :param upper_limit:
            The upper wavelengths limit or None if no upper limit should be applied.
            Default is None.
        :type upper_limit: None, float, Quantity
        :param curve_fit_parameters: Additional parameters for the curve_fit function
        :return: The results of the fit's
        :rtype: DataFrame
        """
        wavelengths, d, err = self._get_sed_data()

        # apply wavelength cuts
        wavelengths, d, err = _apply_wavelength_limits(wavelengths, d, err,
                                                       lower_limit, upper_limit)

        wavelengths, d, err = _rescale(wavelengths, d, err, x_log, y_log)

        d, err = _normalize(d, err, norm)

        colnames = inspect.getfullargspec(func)[0][1:]
        if len(colnames) == 0:
            if 'p0' in curve_fit_parameters:
                # if no names are given, create a list with a, b, c, ... from ascii values
                colnames = [chr(97+i) for i in range(len(curve_fit_parameters['p0']))]
            else:
                raise AttributeError('Unknown number of parameters.\n'
                                     'Use a, b, c, ... instead *parameters or specify p0.')

        if error_weighted:
            fits = []
            for row, row_err in zip(d, err):
                try:
                    m = (row > -9999)
                    if np.sum(m) == 0:
                        fits.append(len(colnames) * [-99999])
                        continue
                    fit, cov = curve_fit(func, wavelengths[m], row[m], sigma=row_err[m],
                                         **curve_fit_parameters)
                    fits.append(fit)
                except RuntimeError:
                    fits.append(len(colnames)*[-99999])
                except TypeError:
                    fits.append(len(colnames)*[-99999])
        else:
            fits = []
            for row in d:
                try:
                    m = (row > -9999)
                    if np.sum(m) == 0:
                        fits.append(len(colnames) * [-99999])
                        continue
                    fit, cov = curve_fit(func, wavelengths[m], row[m],
                                         **curve_fit_parameters)
                    fits.append(fit)
                except RuntimeError:
                    fits.append(len(colnames)*[-99999])
                except TypeError:
                    fits.append(len(colnames)*[-99999])

        self._fits = DataFrame(data=np.array(fits),
                               columns=colnames,
                               index=self.get_index())

        return self._fits

    def create_decomposition(self, n_components):
        if self._fits is None:
            raise AttributeError('No fitting date found. Call first a fitting method.')
        
    def decomposition(self, n_components, model):
        if self._fits is None:
            raise AttributeError('No fitting date found. Call first a fitting method.')

    def fit_blackbody(self, error_weighted=True,
                      lower_limit=None, upper_limit=None):
        """
        Fit's a blackbody on the SED data

        :param error_weighted:
            True, if the fluxes should be weighted by their errors, else False.
            Default is True
        :type error_weighted: bool
        :param lower_limit: The lower limit of the wavelengths. Default is None.
        :type lower_limit: float
        :param upper_limit: The upper limit of the wavelengths. Default is None.
        :type upper_limit: float
        :return:
            The result of the fit's and their errors in the same order as the fluxes.
            The first parameter is the temperature of the black body and the second parameter is
            the rescaling factor after the normalization.
        :rtype: pandas.DataFrame
        """
        wavelengths, d, err = self._get_sed_data()

        # apply wavelength cuts
        wavelengths, d, err = _apply_wavelength_limits(wavelengths, d, err,
                                                       lower_limit, upper_limit)

        # convert angstrom to meter
        wavelengths /= 1E10

        results = []

        if error_weighted:
            for sed, sed_err in zip(d, err):
                try:
                    popt, pcov = curve_fit(_blackbody, wavelengths, sed, sigma=sed_err,
                                           p0=(5000., 1.))
                    results.append(np.append(popt, np.sqrt(np.diag(pcov))))
                except TypeError as e:
                    results.append([-9999, -9999, -9999, -9999])
        else:
            for sed in d:
                try:
                    popt, pcov = curve_fit(_blackbody, wavelengths, sed,
                                           p0=(5000., 1.))
                    results.append(np.append(popt, np.sqrt(np.diag(pcov))))
                except TypeError:
                    results.append([-9999, -9999, -9999, -9999])
        self._fits = DataFrame(np.array(results), columns=['temperature', 'scaling',
                                                           'temperature_err', 'scaling_err'])
        return self._fits

    def __str__(self):
        return str(self._data)

    @property
    def fit(self):
        return self._fits

    @property
    def survey_head(self):
        return self._survey



import pylab as pl
import seaborn
import numpy as np


def replace_labels(axes, cols, labels):
    """
    Sets the the labels of the graph in a proper format. For example removes 'mag' from the old labels

    :param axes: The axes object of the subplot

    :param cols: The names of the cols
    :type cols: list
    :param labels: The new labels
    :type labels: list, dict
    :return:
    """
    if type(labels) == list:
        axes.set_xlabel(labels[0])
        axes.set_ylabel(labels[1])
    elif type(labels) == dict:
        axes.set_xlabel(labels[cols[0]])
        axes.set_ylabel(labels[cols[1]])
    else:
        axes.set_xlabel(cols[0].replace('mag', ''))
        axes.set_ylabel(cols[1].replace('mag', ''))


def create_color_name(c):
    """
    Checks if the color name is a proper color name. If not it will convert the name to a proper style.
    :param c: The input color name
    :type c: str
    :return: The input color name in a proper style
    :rtype: str
    """
    if 'mag' not in c:
        c = c.split('-')
        c = '{}mag - {}mag'.format(c[0].replace(' ', ''),
                                   c[1].replace(' ', ''))
    return c


class Color:
    """
    Class to handle different plot with the colors
    """

    _color = None
    _magnitude = None

    def __init__(self, color, magnitude=None):
        self._color = color
        self._magnitude = magnitude

    def __color_color_multi__(self, cols, labels, legend=False):
        """
        Creates a color-color plot with multiple colors (more than 2) in a multiple grid

        :param cols:
            The name of color columns. Default is None, which means that all colors are taken.
        :type cols: list
        :param labels:
            Replacement of the default labels in a list or a dict. Default is None, which means that default labels
            are used.
        :type labels: list, dict
        :param legend: True if a legend with the selection labels should be shown, else False. Default is False.
        :type legend: bool
        :return:
        """
        # exclude data with nan values
        d = self._color.get_columns(cols)

        m = d[cols[0]] > -999
        print(d.columns)
        print(cols)
        for i in range(1, len(cols)):
            m = m & (d[cols[i]] > -999)

        hue = None
        if self._color.mask.get_mask_count() > 0:
            d['selection'] = '          '
            hue = 'selection'
            for i in range(1, self._color.mask.get_mask_count()):
                d.loc[self._color.mask.get_mask(i)]['selection'] = self._color.mask.get_description(i)
        d = d[m]
        pp = seaborn.PairGrid(d, hue=hue)
        pp.map_diag(pl.hist)
        pp.map_lower(pl.scatter, marker='.')

        pp.fig.subplots_adjust(wspace=0.02, hspace=0.02)

        # change the current labels, which are the default column names to a proper style
        # like removing 'mag' from the labels or replacing the labels with given labels
        for i in range(len(cols)):
            for j in range(len(cols)):
                axes = pp.axes[i][j]
                xlabel = axes.get_xlabel()
                ylabel = axes.get_ylabel()
                if type(labels) == list:
                    if xlabel != '':
                        axes.set_xlabel(labels[i])
                    if ylabel != '':
                        axes.set_ylabel(labels[j])
                elif type(labels) == dict:
                    if xlabel != '':
                        axes.set_xlabel(labels[xlabel])
                    if ylabel != '':
                        axes.set_ylabel(labels[ylabel])
                else:
                    if xlabel != '':
                        axes.set_xlabel(xlabel.replace('mag', ''))
                    if ylabel != '':
                        axes.set_ylabel(ylabel.replace('mag', ''))
        if hue is not None and legend:
            pp.add_legend()
            # pp.fig.tight_layout()

    def __color_color_survey__(self, survey, labels, legend=False):
        d = self._color.get_survey_data(survey)
        m = np.array(len(d)*[True])
        for c in d.columns:
            m = m & (d[c] > -999)

        cols = d.columns.values

        hue = None
        if self._color.mask.get_mask_count() > 0:
            d['selection'] = '          '
            hue = 'selection'
            for i in range(1, self._color.mask.get_mask_count()):
                d.loc[self._color.mask.get_mask(i)]['selection'] = self._color.mask.get_description(i)
        d = d[m]
        pp = seaborn.PairGrid(d, hue=hue)
        pp.map_diag(pl.hist)
        pp.map_lower(pl.scatter, marker='.')

        pp.fig.subplots_adjust(wspace=0.02, hspace=0.02)

        # change the current labels, which are the default column names to a proper style
        # like removing 'mag' from the labels or replacing the labels with given labels
        for i in range(len(cols)):
            for j in range(len(cols)):
                axes = pp.axes[i][j]
                xlabel = axes.get_xlabel()
                ylabel = axes.get_ylabel()
                if type(labels) == list:
                    if xlabel != '':
                        axes.set_xlabel(labels[i])
                    if ylabel != '':
                        axes.set_ylabel(labels[j])
                elif type(labels) == dict:
                    if xlabel != '':
                        axes.set_xlabel(labels[xlabel])
                    if ylabel != '':
                        axes.set_ylabel(labels[ylabel])
                else:
                    if xlabel != '':
                        axes.set_xlabel(xlabel.replace('mag', ''))
                    if ylabel != '':
                        axes.set_ylabel(ylabel.replace('mag', ''))
        if hue is not None and legend:
            pp.add_legend()
            # pp.fig.tight_layout()
        pass

    def __color_color_single__(self, cols, labels, legend=False):
        """
        Creates a color-color plot with two magnitudes

        :param cols:
            The name of color columns. Default is None, which means that all colors are taken.
        :type cols: list
        :param labels:
            Replacement of the default labels in a list or a dict. Default is None, which means that default labels
            are used.
        :type labels: list, dict
        :param legend: True if a legend with the selection labels should be shown, else False. Default is False.
        :type legend: bool
        :return:
        """
        color1 = self._color.get_column(cols[0])
        color2 = self._color.get_column(cols[1])
        sp = pl.subplot()
        sp.scatter(color1,
                   color2,
                   marker='.', c='k')

        # iterate over all masks
        for i in range(self._color.mask.get_mask_count()):
            m = self._color.mask.get_mask(i)
            sp.scatter(color1[m],
                       color2[m],
                       marker='.', label=self._color.mask.get_description(i))
        replace_labels(sp, cols, labels)

        if legend and self._color.mask.get_mask_count() > 0:
            pl.legend(loc='best')

    def color_color(self, survey=None, cols=None, path='', labels=None, legend=False):
        """
        Plots a color color diagram. If their are more than two columns, it will
        plot the color-color diagrams in a grid

        :param cols:
            The name of color columns. Default is None, which means that all colors are taken.
        :type cols: list, dict
        :param path:
            Path to the save place. Default is an empty string, which means that the figure will be shown, only.
        :type path: str
        :param labels:
            Replacement of the default labels in a list or a dict. Default is None, which means that default labels
            are used.
        :type labels: list, dict
        :param survey: A name of a loaded survey to make a color-color plot with the colors of this survey.
        :type survey: str
        :param legend: True if a legend with the selection labels should be shown, else False. Default is False.
        :type legend: bool
        :return:
        """
        # todo: if a survey and cols are set, check if the cols are in the survey colors and then use this colors only
        if cols is None:
            if survey is None:
                cols = self._color.data.columns.values
            else:
                self.__color_color_survey__(survey, labels, legend=legend)
                return
                # cols = self._color.survey_colors[survey].values
        else:
            use_cols = []
            if type(cols) is list:
                for c in cols:
                    use_cols.append(create_color_name(c))
            elif type(cols) == dict:
                # todo: implement a col selection with a dict (maybe a list of dicts?)
                raise NotImplementedError()
            else:
                raise ValueError(f'{type(cols)} is not supported format for color names.')
            cols = use_cols

        pl.clf()
        if len(cols) > 2:
            self.__color_color_multi__(cols, labels, legend=legend)
        else:
            self.__color_color_single__(cols, labels, legend=legend)
        if path != '':
            pl.savefig(path)
        pl.show()

    def color_hist(self, cols=None, bins='auto', histtype='step', range=None, density=False, path='', labels=None):
        """
        Plots a histogram of the color(s).

        :param cols:
            The name of color columns. Default is None, which means that all colors are taken.
        :type cols: list
        :param bins: The number of bins in the histogram. Default is 'auto'.
        :type bins: int, str
        :param histtype: The type of the histogram. Default is 'step'.
        :type histtype: str
        :param range: The range of the x_axis. Default is None.
        :type range: list, tuple
        :param density: True if the histogram should be a density histogram, else False. Default is False.
        :type density: bool
        :param path:
            Path to the save place. Default is an empty string, which means that the figure will be shown, only.
        :type path: str
        :param labels:
            Replacement of the default labels in a list or a dict. Default is None, which means that default labels
            are used.
        :type labels: list, dict
        :return:
        """

        # exclude data with nan values
        m = self._color.data[cols[0]] > -999
        for i in range(1, len(cols)):
            m = m & (self._color.data[cols[i]] > -999)

        pl.clf()
        sp = pl.subplot()
        if type(cols) == list:
            for c in cols:
                sp.hist(self._color.data[c][m], bins=bins, histtype=histtype, range=range, density=density)
        else:
            sp.hist(self._color.data[cols][m], bins=bins, histtype=histtype, range=range, density=density)

        replace_labels(sp, cols, labels)
        if path != '':
            pl.savefig(path)
        pl.show()



from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u
from astroquery.irsa import Irsa
import pandas as pd
import numpy as np

ZWICKY_URL = 'https://irsa.ipac.caltech.edu/cgi-bin/ZTF/nph_light_curves?' \
             'POS=CIRCLE{:+09.4f}{:+09.4f}{:+07.4f}&BANDNAME=r&NOBS_MIN=3&BAD_CATFLAGS_MASK=32768&FORMAT=ipac_table'


def progress_bar(iteration, total, prefix='',
                 suffix='', length=100, fill='█', print_end="\r"):
    percent = f'{100*iteration/total:04.1f}'
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end=print_end)
    # Print New Line on Complete
    if iteration == total:
        print()


def download_light_curve(ra, dec, radius=None, index=None):
    """
    Downloads the light curve from the `Zwicky-survey <https://www.ztf.caltech.edu/>`_ DR1.1

    :param ra: RA coordinate
    :type ra: float, Union
    :param dec: Dec coordinate
    :type dec: float, Union
    :param radius:
        The search radius. If no units are given, the search radius unit is set to arcsec.
        Default is 2 arcsec.
    :type radius: float, Quantity
    :param index:
        The index of the coordinates, if more than one coordinate set is given.
        Otherwise index will be ignored.
    :return: The light curve(s) from PTF
    :rtype: DataFrame
    """
    if radius is None:
        radius = 0.0014

    if type(radius) == u.Quantity:
        radius = radius.to(u.deg)
        radius = radius.value

    if not isinstance(ra, np.float):
        if index is None:
            index = np.arange(len(ra))
        out = []
        counter = 0
        print('Download ZTF light curves')
        for r, d, i in zip(ra, dec, index):
            progress_bar(counter, len(ra), 'Download', '')
            try:
                t = download_light_curve(r, d, radius)
                t['row_id'] = i
                out.append(t)
            except:
                pass
            counter += 1
        progress_bar(counter, len(ra), 'Download', 'Complete')
        return pd.concat(out)

    t = Table.read(ZWICKY_URL.format(ra, dec, radius), format='ascii.ipac')

    return t[['mjd', 'oid', 'mag', 'magerr', 'ra', 'dec']].to_pandas()


def download_ptf(ra, dec, radius=None, index=None):
    """
    Downloads the `PTF <https://www.ptf.caltech.edu/>`_ light curve and reformat the columns
    to align the names to match the naming of the other light curves

    :param ra: RA coordinate
    :type ra: float, Union
    :param dec: Dec coordinate
    :type dec: float, Union
    :param radius:
        The search radius. If no units are given, the search radius unit is set to arcsec.
        Default is 2 arcsec.
    :type radius: float, Quantity
    :param index:
        The index of the coordinates, if more than one coordinate set is given.
        Otherwise index will be ignored.
    :return: The light curve(s) from PTF
    :rtype: DataFrame
    """
    # if no radius is given
    if radius is None:
        radius = 2*u.arcsec
    # if the radius is just a float
    elif type(radius) == float:
        radius = radius*u.arcsec

    if not isinstance(ra, np.float):
        if index is None:
            index = np.arange(len(ra))
        out = []
        counter = 0
        print('Download PTF light curves')
        for r, d, i in zip(ra, dec, index):
            progress_bar(counter, len(ra), 'Download', 'Complete')
            t = download_ptf(r, d, radius, i)
            t['row_id'] = i
            out.append(t)
            counter += 1
        progress_bar(counter, len(ra), 'Download', 'Complete')
        return pd.concat(out)
    s = SkyCoord(ra*u.deg, dec*u.deg)
    table = Irsa.query_region(s, catalog="ptf_lightcurves", spatial="Cone",
                              radius=radius)
    table = table[['obsmjd', 'mag_autocorr', 'magerr_auto', 'ra', 'dec', 'oid']].to_pandas()
    table = table.rename({'obsmjd': 'mjd', 'mag_autocorr': 'mag',
                          'magerr_auto': 'magerr'}, axis='columns')
    return table



import numpy as np


class Fit:
    _x_log = False
    _y_log = False
    _func = None
    _fit_value = None

    def __init__(self, func, x_log=False, y_log=False):
        self.func = func
        self.x_log = x_log
        self.y_log = y_log

    def __call__(self, *args, **kwargs):
        x = args[0]
        y = args[1]
        if self.x_log:
            x = np.log10(x)

        if self.y_log:
            y = np.log10(y)

        fit = self.func(x, y)
        self._fit_value = fit
        return fit

    @property
    def x_log(self):
        return self._x_log

    @x_log.setter
    def x_log(self, value):
        if type(value) == bool:
            self._x_log = value
        else:
            raise ValueError('x_log must be a boolean')

    @property
    def y_log(self):
        return self._y_log

    @y_log.setter
    def y_log(self, value):
        if type(value) == bool:
            self._y_log = value
        else:
            raise ValueError('y_log must be a boolean')

    @property
    def func(self):
        return self._func

    @func.setter
    def func(self, f):
        if callable(f):
            self._func = f
        else:
            raise ValueError('Parameter must be callable function.')

    @property
    def fit_value(self):
        return self._fit_value

    @fit_value.setter
    def fit_value(self, value):
        raise AttributeError('Attribute can not be set')


def poly(degree):
    """
    Returns numpy-polynomial fit of the given degree

    :param degree: The degree of the polynomial
    :type degree: int
    :return:
    """
    def poly_fit(x, y):
        return np.polyfit(x, y, degree)
    return poly_fit


class PolyFit(Fit):

    def __init__(self, degree, **kwargs):
        Fit.__init__(self, poly(degree), **kwargs)



from astropy.table import Table, vstack
from astropy.modeling import models, fitting
from astropy import units as u
import numpy as np
import glob
import os
import numbers

from Phosphorpy.data.sub.plots.spectra import SpectraPlot, SpectraListPlot


class SpectraList:

    _spectra = None
    _ids = None
    _plot = None

    def __init__(self, spectra=None, ids=None):
        """
        SpectraList provides the handling of multiple spectra

        :param spectra: Optional: A spectra or a list of spectra
        :type spectra: None, Spectra, list
        :param ids:
            Optional: ID's of the spectra. If they are given, they must have the same length as the given spectra.
        :type ids: int, list
        """
        if spectra is None:
            self._spectra = []
            self._ids = []
        elif type(spectra) == list or type(spectra) == tuple:
            self._spectra = list(spectra)
            if ids is None:
                self._ids = list(np.arange(len(spectra)))
            else:
                if len(spectra) == len(ids):
                    self._ids = list(ids)
                else:
                    raise ValueError('If spectra and ids are given, they must have the same length.')
        else:
            self._spectra = [spectra]
            if ids is not None:
                self._ids = [ids]
            else:
                self._ids = [0]

        self._plot = SpectraListPlot(self)

    def __len__(self):
        return len(self._spectra)

    def __str__(self):
        return f'SpectraList with {len(self)} spectra.'

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        """

        :param item:
        :return:
        :rtype: Spectra, int
        """
        if isinstance(item, numbers.Integral):
            return self._spectra[item], self._ids[item]
        elif type(item) == list or type(item) == tuple or type(item) == np.ndarray:
            spec_out = [
                self._spectra[i] for i in item
            ]
            ids_out = [
                self._ids[i] for i in item
            ]
            return SpectraList(spec_out, ids_out)

    def write(self, path, data_format='fits', overwrite=True):
        """
        Writes the spectra of this SpectraList to files

        :param path: The path to the directory, where the files should be placed.
        :type path: str
        :param data_format: The format of the spectra files
        :type data_format: str
        :param overwrite: True, if an existing file should be overwritten, else False. Default is True
        :type overwrite: bool
        :return:
        """
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise ValueError('path must be a directory and not a file.')

        paths = []
        for spec, index in zip(self._spectra, self._ids):
            paths.append(spec.write(
                f'{path}{spec.survey}_{index}.{data_format}',
                data_format=data_format, overwrite=overwrite
            ))
        return paths

    @staticmethod
    def read(path, format='fits', wavelength_name='wavelength', flux_name='flux', survey_key='survey'):
        """
        Reads all spectra from the directory

        :param path: Path to the directory with the spectra
        :type path: str
        :param format: The format of the files of the spectra.
        :type format: str
        :param wavelength_name: The name of the wavelength column
        :type wavelength_name: str
        :param flux_name: The name of the flux column
        :type flux_name: str
        :param survey_key: The key of the survey information
        :type survey_key: str
        :return: The SpectraList with all the red spectra
        :rtype: SpectraList
        """
        spec_list = SpectraList()
        for f in glob.glob(f'{path}*.{format}'):
            spec = Spectra.read(f, format=format,
                                wavelength_name=wavelength_name, flux_name=flux_name,
                                survey_key=survey_key)
            spec_list.append(spec)
        return spec_list

    def get_by_id(self, index):
        """
        Returns the spectra with the required index

        :param index: The index of the required spectra
        :type index: int
        :return:
            SpectraList with the required spectra. The list can contain more than one spectra, if multiple spectra
            with the required ID are found.
            If no spectra is found with the required ID, the method will return None.
        :rtype: SpectraList, None
        """
        if isinstance(index, numbers.Integral):
            con = index in self._ids
        else:
            con = np.isin(index, np.array(self._ids)).any()

        if con:
            p = np.where(np.array(self._ids) == index)[0]
            spec_list = SpectraList()
            for i in p:
                spec_list.append(*(self[i]))
            return spec_list

    def get_ids(self):
        """
        Returns all IDs of the downloaded spectra. Every ID is returned only once but multiple spectra could
        be available.
        :return: The IDs of the spectra
        :rtype: ndarray
        """
        return np.unique(self._ids)

    def append(self, spectra, spec_id=-1):
        """
        Appends a new spectra to the spectra list.

        :param spectra: The new spectra
        :type spectra: LamostSpectra
        :param spec_id: The ID of the spectra
        :type spec_id: int
        :return:
        """
        self._spectra.append(spectra)
        if spec_id > -1:
            self._ids.append(spec_id)
        else:
            self._ids.append(len(self._ids))

        if spectra.index == -1:
            spectra.index = self._ids[-1]

    def merge(self, second):
        """
        Merges a second SpectraList into this one

        :param second: The other SpectraList
        :type second: SpectraList
        :return:
        """
        if second is None:
            return

        if type(second) != SpectraList:
            raise ValueError('second must be of the type \'SpectraList\'')

        for i in range(len(second)):
            temp = second[i]
            self.append(temp[0].copy(), temp[1])

    # def estimate_line_properties(self, as_velocity=False, redo=False):
    #     """
    #     Estimates the line properties of all spectra in the list
    #
    #     :param as_velocity:
    #         True, if the line shift should be returned as a radial velocity in km/s, else False to
    #         get lambda/lambda0-1.
    #         Default is False.
    #     :type as_velocity: bool
    #     :param redo:
    #         True, if old results should be ignored, else False.
    #         Default is False.
    #     :type redo: bool
    #     :return: The line properties of all spectra
    #     :rtype: dict
    #     """
    #     out = {}
    #     for s in self._spectra:
    #         out[s.obs_id] = s.estimate_line_properties(as_velocity=as_velocity,
    #                                                    redo=redo)
    #     return out
    #
    # def as_dataframe(self, as_velocity=False, redo=False):
    #     """
    #     Returns the main information of the stored spectra as a pandas DataFrame
    #
    #     :param as_velocity:
    #     :param redo:
    #     :return: The main information as a dataframe
    #     :rtype: DataFrame
    #     """
    #     out = []
    #     for s, d_id in zip(self._spectra, self._ids):
    #         properties = s.estimate_line_properties(as_velocity=as_velocity,
    #                                                 redo=redo)
    #         properties['obsID'] = s.obs_id
    #         properties['ID'] = d_id
    #         out.append(properties)
    #     out = vstack(out).to_pandas()
    #     return out.set_index('ID')

    @property
    def plot(self):
        return self._plot


class Spectra:
    NORMALIZE_MEAN = np.mean
    """
    Static variable for the normalization with the mean value
    """
    NORMALIZE_MEDIAN = np.median
    """
    Static variable for the normalization with the median value
    """
    NORMALIZE_MAX = np.max
    """
    Static variable for the normalization with the max value
    """
    NORMALIZE_SUM = np.sum
    """
    Static variable for the normalization with the sum
    """

    _wavelength = None
    _flux = None

    _wavelength_unit = None
    _flux_unit = None

    _survey = None
    _index = -1
    _meta = None

    _fits = None

    _lines = None

    _plot = None

    def __init__(self, wavelength=None, flux=None, wavelength_unit=None, flux_unit=None, survey=None,
                 index=-1, meta=None):
        """
        Spectra is the basic class to handle different kinds of spectra on a basic level
        without any specific functionality related to any certain spectra

        :param wavelength:
            The wavelength values of the spectra or None.
            As wavelength values a Union (tuple, list, array) is allowed or a Quantity.
        :type wavelength: Union, Quantity
        :param flux:
            The flux values of the spectra or None.
            As flux values a Union (tuple, list, array) is allowed or a Quantity.
        :type flux: Union, Quantity
        :param wavelength_unit:
            The units of the wavelengths, if the wavelengths are given. If no unit is given, angstrom are
            assumed to be the wavelength unit.
        :type wavelength_unit: Unit
        :param flux_unit:
            The units of the flux, if the flux is given. If no unit is given, ergs are assumed to be
            the flux unit.
        :type flux_unit: Unit
        :param survey: Name of the survey. Default is None.
        :type survey: str, None
        :param index: Index of the spectra. Default is -1.
        :type index: int
        """
        if type(wavelength) == u.Quantity:
            wavelength_unit = wavelength.unit
            wavelength = wavelength.value

        if type(flux) == u.Quantity:
            flux_unit = flux.unit
            flux = flux.value

        self._wavelength = wavelength
        self._flux = flux

        self.wavelength_unit = wavelength_unit
        self.flux_unit = flux_unit

        self._survey = survey
        self._index = index
        self._meta = meta

        self._plot = SpectraPlot(self)
        self._fits = []

    def __str__(self):
        return f'Spectra with wavelength between {self.min_wavelength} and {self.max_wavelength} with a' \
               f' wavelength resolution of {np.round(self.resolution_wavelength, 2)}.'

    def __repr__(self):
        return str(self)

    def write(self, path, data_format='fits', overwrite=True):
        """
        Writes the spectral data to a file

        :param path: The path to the file.
        :type path: str
        :param data_format: The format of the output file. Current supported formats are fits and csv.
        :type data_format: str
        :param overwrite:
            True if an existing file should be overwritten, else False.
            Default is False.
        :return:
        """
        tab = Table()
        tab['wavelength'] = self.wavelength
        tab['wavelength'].unit = self.wavelength_unit
        tab['flux'] = self.flux
        tab['flux'].unit = self.flux_unit
        tab.meta['survey'] = self.survey
        tab.meta['index'] = self._index

        if data_format == 'fits':
            data_format = 'fits'
        elif data_format == 'csv':
            data_format = 'ascii.csv'
        else:
            raise ValueError(f'Format: {format} is not supported.\nChoose \'fits\' or \'csv\'.')

        tab.write(path, format=data_format, overwrite=overwrite)
        return path

    @staticmethod
    def read(path, format='fits', wavelength_name='wavelength', flux_name='flux', survey_key='survey'):
        """
        Reads a spectra from a file

        :param path: Path to the file.
        :type path: str
        :param format: Format of the file. Current supported formats are fits and csv
        :type format: str
        :param wavelength_name: The name of the wavelength column
        :type wavelength_name: str
        :param flux_name: The name of the flux column
        :type flux_name: str
        :param survey_key: The name of the header entry with the name of the survey/origin
        :type survey_key: str
        :return: The Spectra-object with the data from the file
        :rtype: Spectra
        """
        if format == 'fits':
            tab = Table.read(path, format='fits')
        elif format == 'csv':
            tab = Table.read(path, format='ascii.csv')
        else:
            raise ValueError(f'Format: {format} is not supported.\nChoose \'fits\' or \'csv\'')

        try:
            survey = tab.meta[survey_key]
        except KeyError:
            survey = None

        try:
            index = tab.meta['index']
        except KeyError:
            index = -1
        return Spectra(
            wavelength=tab[wavelength_name],
            flux=tab[flux_name],
            wavelength_unit=tab[wavelength_name].unit,
            flux_unit=tab[flux_name].unit,
            survey=survey, index=index
        )

    def copy(self):
        """
        Make a copy of the Spectra class

        :return: A copy of the current Spectra-object
        :rtype: Spectra
        """
        return Spectra(
            wavelength=self.wavelength.copy(),
            flux=self.flux.copy(),
            wavelength_unit=self.wavelength_unit,
            flux_unit=self.flux_unit,
            survey=self.survey, index=self.index
        )

    def normalize(self, kind, inplace=True):
        """
        Normalize the spectra with a certain function.

        .. math::

            flux^* = \frac{flux}{function(flux)}


        :param kind:
            The kind of function, which is used to normalize the spectra.
            Allowed function are the static function or 'mean', 'median', 'max' or 'sum' as ky words.
        :type kind: function, str
        :param inplace:
            True, if the flux should be overwritten with the normalized flux, else False for a new spectra
            object. Default is True.
        :type inplace: bool
        :return: None or if inplace is False, a new spectra object.
        :rtype: None, Spectra
        """
        if type(kind) == str:
            kind = kind.lower()
        fl = self.flux.copy()
        if kind == self.NORMALIZE_MAX or kind == 'max':
            norm = np.max(fl)
        elif kind == self.NORMALIZE_MEAN or kind == 'mean':
            norm = np.mean(fl)
        elif kind == self.NORMALIZE_MEDIAN or kind == 'median':
            norm = np.median(fl)
        elif kind == self.NORMALIZE_SUM or kind == 'sum':
            norm = np.sum(fl)
        else:
            raise ValueError('Kind of normalization is not allowed.')

        if norm == 0:
            raise ValueError('Normalization not possible. Norm is 0.')

        fl /= norm

        if type(inplace) == bool:
            if inplace:
                self._flux = fl
                self._flux_unit /= self._flux_unit
            else:
                return Spectra(wavelength=self.wavelength.copy(),
                               flux=fl)
        else:
            raise TypeError('Inplace must be a bool-type.')

    def cut(self, min_wavelength=None, max_wavelength=None, inplace=True):
        """
        Make a cutout of the spectra between the minimal and maximal wavelengths.

        :param min_wavelength:
            The minimal wavelength as a float, then the we assume the same units as the wavelength,
            a Quantity with the dimension length or None.
        :type min_wavelength: float, Quantity, None
        :param max_wavelength:
            The maximal wavelength as a float, then the we assume the same units as the wavelength,
            a Quantity with the dimension length or None.
        :type max_wavelength: float, Quantity, None
        :param inplace:
            True, if the current spectra should be overwritten with the cutout, else False to get
            a new Spectra object. Default is True.
        :type inplace: bool
        :return:
        """
        # if both are None, raise an error because it does not make any sense
        if min_wavelength is max_wavelength:
            raise ValueError('At least one of minimal wavelength or maximal wavelength must be given.')

        wave = self.wavelength.copy()
        flux = self.flux.copy()

        if min_wavelength is not None:
            # if the minimal wavelength is a astropy Quantity, align the units and apply the limit then
            if type(min_wavelength) == u.Quantity:
                min_wavelength = min_wavelength.to(self.wavelength_unit).value
            m = wave >= min_wavelength
            flux = flux[m]
            wave = wave[m]

        if max_wavelength is not None:
            # if the maximal wavelength is a astropy Quantity, align the units and apply the limit then
            if type(max_wavelength) == u.Quantity:
                max_wavelength = max_wavelength.to(self.wavelength_unit).value
            m = wave <= max_wavelength
            flux = flux[m]
            wave = wave[m]

        if type(inplace) == bool:
            if inplace:
                self._wavelength = wave
                self._flux = flux
            else:
                return Spectra(wavelength=wave, flux=flux,
                               wavelength_unit=self.wavelength_unit,
                               flux_unit=self.flux_unit)
        else:
            raise ValueError('inplace must be a bool.')

    def fit_line(self, model=None):
        """
        Fit's the spectra/line with the model. The model must be a fittable object from
        the astropy.modeling.model classes.
        If no model is given, a 1D gaussian is used.

        :param model: The fittable model or None. Default is None, which means that a 1D gaussian is used.
        :return: The resulting fit
        """
        if model is None:
            model = models.Gaussian1D(stddev=10)
        else:
            try:
                model.fittable
            except AttributeError:
                raise ValueError(f'Model must be an instance of astropy\'s Fittable1DModel and not {type(model)}')

        fitter = fitting.SLSQPLSQFitter()
        fit_rs = fitter(model, self.wavelength, self.flux)
        self._fits.append(fit_rs)
        return fit_rs

    def fit_gauss(self, guesses):
        """
        Fit's the current spectra/line with a 1D gaussian

        :param guesses: Initial guesses for the gaussian
        :type guesses: dict
        :return: The resulting fit
        """
        gauss = models.Gaussian1D(**guesses)
        return self.fit_line(model=gauss)

    def fit_double_gauss(self, guesses, guesses2):
        """
        Fit's the current spectra/line(s) with two 1D gaussian's

        :param guesses: Initial guesses for the first gaussian
        :type guesses: dict
        :param guesses2: Initial guesses for the second gaussian
        :type guesses2: dict
        :return: The resulting fit
        """
        dgauss = models.Gaussian1D(**guesses)+models.Gaussian1D(**guesses2)
        return self.fit_line(model=dgauss)

    def estimate_line_properties(self):
        pass

    @property
    def wavelength(self):
        """
        The wavelengths of spectra
        :return:
        """
        return self._wavelength.copy()

    @property
    def wavelength_unit(self):
        """
        The wavelength unit
        :return:
        """
        return self._wavelength_unit

    @wavelength_unit.setter
    def wavelength_unit(self, unit):
        """
        Sets a new wavelength unit
        :param unit: The unit of the dimension length
        :type unit: Unit
        :return:
        """
        # if no unit is given, assume angstrom
        if unit is None:
            # if no previous unit was set
            if self._wavelength_unit is None:
                self._wavelength_unit = u.angstrom
            # if wavelength has a unit, recursive call to change the unit to angstrom
            else:
                self.wavelength_unit = u.angstrom
        elif not isinstance(unit, u.Unit):
            raise ValueError('The new unit must be a astropy unit.')
        elif self._wavelength_unit is None:
            self._wavelength_unit = unit
        else:
            try:
                # estimate the conversion factor and apply it to the wavelengths
                factor = (self._wavelength_unit.to(unit))
                self._wavelength *= factor
                self._wavelength_unit = unit
            except u.UnitConversionError:
                raise ValueError('Unit must be equivalent to a length.')

    @property
    def flux(self):
        """
        The flux values
        :return:
        """
        return self._flux

    @property
    def flux_unit(self):
        """
        The flux unit
        :return:
        """
        return self._flux_unit

    @flux_unit.setter
    def flux_unit(self, unit):
        """
        Sets a new flux unit
        :param unit: The unit of the dimension of a flux
        :type unit: Unit
        :return:
        """
        if unit is None:
            self._flux_unit = u.erg
        elif not isinstance(unit, u.Unit):
            raise ValueError('The new unit must be a astropy unit.')
        elif self._flux_unit is None:
            self._flux_unit = unit
        else:
            try:
                # estimate the conversion factor and apply it to the fluxes
                factor = (self._flux_unit.to(unit))
                self._flux *= factor
                self._flux_unit = unit
            except u.UnitConversionError:
                raise ValueError('Unit must be equivalent to a flux.')

    @property
    def plot(self):
        """
        The plotting environment
        :return:
        """
        return self._plot

    @property
    def min_wavelength(self):
        return self.wavelength.min()

    @property
    def max_wavelength(self):
        return self.wavelength.max()

    @property
    def resolution_wavelength(self):
        return np.mean(self._wavelength[1:]-self._wavelength[:-1])

    @property
    def min_flux(self):
        return self.flux.min()

    @property
    def max_flux(self):
        return self.flux.max()

    @property
    def fit(self):
        return self._fits

    @property
    def survey(self):
        return self._survey

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        self._index = value



from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from http.client import RemoteDisconnected

from Phosphorpy.data.sub.colors import Colors
from Phosphorpy.data.sub.coordinates import CoordinateTable
from Phosphorpy.data.sub.flux import FluxTable
from Phosphorpy.data.sub.light_curve import LightCurves
from Phosphorpy.data.sub.magnitudes import MagnitudeTable as Magnitude, SurveyData
from Phosphorpy.data.sub.plots.plot import Plot
from Phosphorpy.data.sub.table import Mask
from Phosphorpy.external.image import PanstarrsImage, SDSSImage
from Phosphorpy.external.spectra import get_spectra, get_all_spectra
from Phosphorpy.external.vizier import query_by_name, query_simbad, constrain_query
from Phosphorpy.data.sub.astrometry import AstrometryTable

try:
    from extinction.extinction import get_extinctions
except ImportError:
    from Phosphorpy.external.extinction import get_extinctions

import pandas as pd
from pandas import DataFrame
import numpy as np
import zipfile
import os
import warnings
import sys


def add_to_zip(zi, data, name, format='csv'):
    """
    Writes a file to zip file

    :param zi: The zip-file object

    :param data: The data file to write
    :type data: Phosphorpy.data.sub.table.DataTable, Phosphorpy.data.sub.magnitudes.Survey
    :param name: Name of the file
    :type name: str
    :param format: The format of the file
    :type format: str
    :return:
    """
    p = f'./{name}'
    paths = data.write(p, data_format=format)
    if paths is None:
        zi.write(p, name)
        os.remove(p)
    else:
        for p in paths:
            zi.write(p, p.split('./')[-1])
            os.remove(p)


def read_from_zip(zi, name):
    """
    Reads a file from a zip archive and converts to a pandas DataFrame

    :param zi: The zip-file object
    :type zi: zipfile.ZipFile
    :param name: Name of the member file
    :type name: str
    :return: The red item
    :rtype: pandas.DataFrame
    """
    zi.extract(name, '.')
    ending = name.split('.')[-1]
    if ending == 'fits':
        d = Table.read(name).to_pandas()
    elif ending == 'csv':
        d = pd.read_csv(name, index_col=0)
    elif ending == 'ini':
        return SurveyData.read(name)
    else:
        raise ValueError(f'Format {format} is not supported.')
    os.remove(name)
    return d


class DataSet:
    _index = None
    _head = None
    _mask = None
    _coordinates = None
    _magnitudes = None
    _colors = None
    _flux = None
    _astrometry = None

    _light_curves = None
    _spectra = None
    _plot = None

    def __init__(self, data=None, index=None, coordinates=None, magnitudes=None, colors=None, flux=None):
        """
        Standard data class for a survey like dataset. It requires a data file or at least coordinates and magnitudes.
        If no other data are given, it will try to compute the colors and index the sources.

        :param data: An input dataset in a numpy.ndarray style
        :type data: numpy.ndarray, astropy.table.Table, pandas.DataFrame
        :param index: A list of index, they must have the same length as coordinates, magnitudes and colors
        :type index: numpy.ndarray
        :param coordinates: A list with coordinates, they must have the same length as index, magnitudes and colors
        :type coordinates: numpy.ndarray
        :param magnitudes: A list with magnitudes, the must have the same length as index, coordinates and colors
        :type magnitudes: numpy.ndarray
        :param colors: A list with colors, they must have the same length as index, coordinates and magnitudes
        :type colors: numpy.ndarray
        """
        try:
            self._mask = Mask(len(data))
        except TypeError:
            self._mask = Mask(len(coordinates))
        if data is not None:
            if type(data) == np.ndarray:
                cols = data.dtype.names
            elif type(data) == Table:
                cols = data.colnames
            elif type(data) == DataFrame:
                cols = data.columns
            else:
                raise TypeError(f'Unsupported data type: {type(data)}')

            if 'index' in cols:
                self._index = np.array(data['index'], dtype=np.int32)
            else:
                self._index = np.linspace(0, len(data), len(data), dtype=np.int32)
            self._coordinates = CoordinateTable(data, mask=self._mask)
            self._magnitudes = Magnitude(data, mask=self._mask)

        # if no data are given but coordinates, magnitudes and maybe colors
        elif coordinates is not None and magnitudes is not None:
            # if coordinates.shape[0] != magnitudes.shape[0]:
            #     raise AttributeError('Coordinates and magnitudes do not have the same length!')
            if type(coordinates) != CoordinateTable:
                coordinates = CoordinateTable(coordinates, mask=self._mask)

            self._coordinates = coordinates

            if type(magnitudes) != Magnitude:
                magnitudes = Magnitude(magnitudes, mask=self._mask)

            self._magnitudes = magnitudes

            if colors is not None:
                if type(colors) is bool and colors:
                    self._colors = self.magnitudes.get_colors()
                else:
                    if type(colors) != Colors:
                        self._colors = Colors(colors, mask=self._mask)
                    else:
                        self._colors = colors

            if flux is not None:
                if type(flux) != FluxTable:
                    flux = FluxTable(flux)

                self._flux = flux

            if index is not None:
                self._index = index
            else:
                self._index = np.arange(1, len(coordinates), 1)
        elif coordinates is not None:
            self._coordinates = CoordinateTable(coordinates, mask=self._mask)
        else:
            raise AttributeError('data or at least coordinates and magnitudes are required!')

        self._plot = Plot(self)

    def __return_masked__(self, attribute):
        """
        Returns the input attribute with a applied mask, if a mask is set.

        :param attribute: One attribute of this class object
        :type attribute: CoordinateTable, Magnitude
        :returns: The attribute with an applied mask, if a mask is set. If no mask is set, the whole attribute
        :rtype: numpy.ndarray
        """
        if self._mask.get_mask_count() > 0:
            return attribute[self._mask.mask]
        else:
            return attribute

    def set_astrometry(self, astrometry):
        self._astrometry = astrometry

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def magnitudes(self):
        """

        :return: MagnitudeTable of the DataSet
        :rtype: Magnitude
        """
        return self._magnitudes

    @property
    def colors(self):
        if self._colors is None:
            self._colors = self._magnitudes.get_colors()
        return self._colors

    @property
    def flux(self):
        if self._flux is None:
            self._flux = self._magnitudes.get_flux()
        return self._flux

    @property
    def index(self):
        return self._index

    @property
    def plot(self):
        return self._plot

    @property
    def astrometry(self):
        if self._astrometry is None:
            self._astrometry = AstrometryTable.load_astrometry(self._coordinates)
        return self._astrometry

    @property
    def spectra(self):
        if self._spectra is None:
            self._spectra = get_all_spectra(
                self.coordinates.to_sky_coord(),
                self.coordinates.data.index.values
            )
        return self._spectra

    def get_spectra(self, coordinate=None, ra=None, dec=None):
        """
        Returns the spectra close to the given coordinates.

        :param coordinate: Coordinates (or close to) of the required ID.
        :type coordinate: SkyCoord
        :param ra: The RA component of the coordinates
        :type ra: float
        :param dec: The Dec component of the coordinates
        :type dec: float
        :return: The closest spectra
        :rtype: SpectraList
        """
        min_id, distance = self.coordinates.get_closest_source_id(coordinate, ra, dec)
        return self.spectra.get_by_id(min_id)

    def get_light_curve(self, coordinate=None, ra=None, dec=None):
        """
        Returns the light curve close to the given coordinates.

        :param coordinate: Coordinates (or close to) of the required ID.
        :type coordinate: SkyCoord
        :param ra: The RA component of the coordinates
        :type ra: float
        :param dec: The Dec component of the coordinates
        :type dec: float
        :return: The closest light curve
        :rtype: LightCurves
        """
        min_id, distance = self.coordinates.get_closest_source_id(coordinate, ra, dec)
        return self.light_curves.get_light_curve(min_id)

    def remove_unmasked_data(self):
        """
        Removes all unmasked items from the dataset and sets the mask back to None.

        :return:
        """
        # self._index = self.__return_masked__(self._index)

        self._coordinates.remove_unmasked_data()
        if self._magnitudes is not None:
            self._magnitudes.remove_unmasked_data()
        if self._colors is not None:
            self._colors.remove_unmasked_data()
        if self._flux is not None:
            self._colors.remove_unmasked_data()
        if self._astrometry is not None:
            self._astrometry.remove_unmasked_data()
        self._coordinates.mask.reset_mask()

    def __get_attribute__(self, item):
        """
        Returns the corresponding attribute to item, if the item doesn't match with
        one of the names, a KeyError will raise.

        :param item:
            The name of the attribute. Allowed strings are 'index', 'coordinate', 'coordinates',
            'magnitude', 'magnitudes', 'color', 'colors' and 'mask'
        :type item: str
        :returns: The corresponding attribute
        :rtype: numpy.ndarray
        """
        if item == 'index':
            return self.index
        elif item == 'coordinates' or item == 'coordinate':
            return self.coordinates
        elif item == 'magnitudes' or item == 'magnitude':
            return self.magnitudes
        elif item == 'colors' or item == 'color':
            return self.colors
        elif item == 'mask':
            return self._mask
        else:
            error = f'Key {item} not known! Possible option are index, coordinates, magnitudes and colors.'
            raise KeyError(error)

    def __get_row__(self, item):
        """
        Returns the values of the data element or row wise
        :param item: A slice or an int to indicate the required data
        :type item: slice, int
        :return:
        """
        if type(item) == slice:
            out = self.coordinates.data[item].merge(self.magnitudes.data[item],
                                                    left_index=True,
                                                    right_index=True)
            if self.colors is not None:
                out = out.merge(self.colors[item],
                                left_index=True,
                                right_index=True)

            return out

        elif type(item) == int:
            out = self.coordinates[item]
            if self._magnitudes.has_data():
                out = out.concat(self.magnitudes[item], axis=1)
            if self._colors is not None and self._colors.has_data():
                out = out.concate(self.colors[item], axis=1)

            if self._flux is not None and self._flux.has_data():
                out = out.concate(self.flux[item], axis=1)
            return out

    def __getitem__(self, item):

        # if the item is a string, return one of the attributes of the object
        if type(item) == str:
            return self.__get_attribute__(item)
        # if the item is a slice or an int, return the corresponding data
        else:
            return self.__get_row__(item)

    def __load_from_vizier__(self, name):
        d = query_by_name(name, self.coordinates.to_table())
        self._magnitudes.add_survey_mags(d, name.lower())
        self._colors = None
        self._flux = None

    def load_from_vizier(self, name):
        """
        Load new photometric data from Vizier

        :param name:
            A single name or a list of names of surveys or one of the keywords 'optical' or 'nir' to
            download all optical or NIR surveys
        :type name: str, list, tuple
        :return:
        """
        if type(name) == str:
            if name.lower() == 'optical':
                self.load_from_vizier(['sdss', 'ps', 'kids', 'gaia'])
            elif name.lower() == 'nir':
                self.load_from_vizier(['2mass', 'ukidss', 'viking'])
            elif name.lower() == 'all':
                self.load_from_vizier(['galex', 'optical', 'nir', 'wise'])
            else:
                self.__load_from_vizier__(name)
        elif type(name) == list or type(name) == tuple:
            for n in name:
                self.load_from_vizier(n)

    def load_spectra(self, survey):
        """
        Loads all available spectra for the set coordinates

        :param survey:
            The name of the survey from which the spectra should be taken.
            Currently, available surveys are 'SDSS', 'LAMOST' and 'GAMA'
        :return:
        """
        specs = get_spectra(self.coordinates.to_sky_coord(), survey)

        if self._spectra is None:
            self._spectra = specs
        else:
            self._spectra.merge(specs)

        return self._spectra

    def add_magnitudes(self, data, survey_info, ra_name='ra', dec_name='dec'):
        """
        Adds magnitudes from a different source to the DataSet

        :param data: The new data with coordinates and the magnitudes
        :type data: pd.DataFrame, Table
        :param survey_info:
        :param ra_name: The name of the R.A. column
        :type ra_name: str
        :param dec_name: The name of the Dec column
        :type dec_name: str
        :return:
        """
        # check the type of data
        if type(data) != (pd.DataFrame or Table):
            raise ValueError('pandas DataFrame or astropy Table is required')
        elif type(data) == Table:
            data = data.to_pandas()

        # rename the columns if they aren't the default names
        if ra_name != 'ra' or dec_name != 'dec':
            data = data.rename({ra_name: 'ra', dec_name: 'dec'})
        # TODO: implement
        data = self.coordinates.match(data)
        pass

    def get_simbad_data(self):
        return query_simbad(self.coordinates)

    def images(self, survey, source_id, directory='', bands=None, size=None, smooth=0):
        """
        Download images from SDSS or Pan-STARRS and create a colored image out of them

        :param survey: SDSS or Pan-STARRS
        :type survey: str
        :param source_id: The id of the source
        :type source_id: int
        :param directory: The path to the directory, where the images should be saved
        :type directory: str
        :param bands: Individual filter combination. Default is None, which means the default combination is used.
        :type bands: None, tuple, list
        :param size: The wanted size of the image
        :type size: float, astropy.units.Quantity
        :param smooth: Number of smooths. Default is 0.
        :type smooth: int
        :return:
        """
        survey = survey.lower()

        # if sdss is the selected survey
        if survey == 'sdss':
            s = SDSSImage()
        # if pan-starrs is the selected survey
        elif survey == 'panstarrs' or survey == 'ps' or survey == 'ps1' or survey == 'pan-starrs':
            s = PanstarrsImage()
        else:
            raise ValueError('{} is not allowed. Only SDSS or PANSTARRS as surveys are allowed.')
        coord = self.coordinates.to_sky_coord(source_id)

        # if the directory string is not empty
        if directory != '':
            # if the last character is not '/'
            if directory[-1] != '/':
                # add '/' to the end of the directory string
                directory = directory + '/'
            directory = directory + coord.to_string('hmsdms')+'.png'
        try:
            s.get_color_image(coord, directory, bands=bands, size=size, smooth=smooth)
        except ValueError:
            warnings.warn("Image is not available", UserWarning)
        except RemoteDisconnected:
            warnings.warn('Connection interrupted.')

    def all_images(self, survey, directory='', bands=None, size=None, smooth=0):
        """
        Downloads images of all sources and create a colored image for every source.

        :param survey: SDSS or Pan-STARRS
        :type survey: str
        :param directory: The path to the directory, where the images should be stored
        :type directory: str
        :param bands: Individual filter combination. Default is None, which means the default combination is used.
        :type bands: None, tuple, list
        :param size: The wanted size of the image
        :type size: float, astropy.units.Quantity
        :param smooth: Number of smoothings. Default is 0.
        :type smooth: int
        :return:
        """
        try:
            m = self._mask.get_latest_mask().values
        except IndexError:
            m = np.ones(len(self.coordinates))
        for i in range(len(self.coordinates)):
            if not m[i]:
                continue
            self.images(survey, i, directory, bands=bands, size=size)

    @property
    def light_curves(self):
        """
        Downloads all light curves from CSS
        :return:
        """
        if self._light_curves is None:
            self._light_curves = LightCurves(self._coordinates)
        return self._light_curves

    def correct_extinction(self):
        """
        Estimates and applies the extinction correction based on
        Schlafly & Finkbeiner 2011 (ApJ 737, 103) to the magnitudes.
        If fluxes are colors are already computed, then they will be computed
        again after the correction.
        :return:
        """
        if self.magnitudes.data is None:
            raise AttributeError('No magnitudes are set.')
        for s in self.magnitudes.survey.get_surveys():
            extinc = get_extinctions(self.coordinates['ra'],
                                     self.coordinates['dec'],
                                     wavelengths=self.magnitudes.survey.get_survey_wavelengths(s),
                                     filter_names=self.magnitudes.survey.get_survey_magnitude(s))
            self.magnitudes.apply_extinction_correction_to_survey(extinc, s)
        # extinc = get_extinctions(self.coordinates['ra'],
        #                          self.coordinates['dec'],
        #                          wavelengths=self.magnitudes.survey.get_all_wavelengths(),
        #                          filter_names=self.magnitudes.survey.all_magnitudes())
        # self.magnitudes.apply_extinction_correction(extinc)

        # recompute flux and colors if they are already computed
        if self._flux is not None:
            self._flux = None
            self.flux

        if self._colors is not None:
            self._colors = None
            self.colors

    def reset_masks(self):
        """
        Resets the mask

        :return:
        """
        if self._magnitudes is not None:
            self.magnitudes.set_mask(self._mask)
        if self._flux is not None:
            self.flux.set_mask(self._mask)
        if self._colors is not None:
            self.colors.set_mask(self._mask)
        self.coordinates.set_mask(self._mask)

    def __combine_all__(self):
        """
        Combines all tables to one big table

        :return: The combined table
        :rtype: pandas.DataFrame
        """
        comb = self.coordinates.to_table()
        if self.magnitudes is not None and self.magnitudes.data is not None:
            comb = comb.join(self.magnitudes.combine(), how='outer')
        if self._flux is not None:
            fluxes = self.flux.combine()
            fluxes = fluxes.rename(
                dict(zip(
                    fluxes.columns.values,
                    [f.replace('mag', 'flux') for f in fluxes.columns.values]
                )),
                axis='columns'
            )

            comb = comb.join(fluxes, how='outer')

        if self._colors is not None:
            comb = comb.join(self.colors.combine(), how='outer')
        return comb

    def __write_as_zip__(self, path, format):
        """
        Creates a zip file with the data of the DataSet

        :param path: Path to the save place
        :type path: str
        :param format: Format of the internal data-files
        :type format: str
        :return:
        """
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        with zipfile.ZipFile(path, 'w', zipfile.ZIP_DEFLATED) as zi:
            add_to_zip(zi, self.coordinates, f'coordinates.{format}', format=format)
            if self._magnitudes.data is not None:
                add_to_zip(zi, self.magnitudes, f'magnitudes.{format}', format=format)
                add_to_zip(zi, self.magnitudes.survey, 'survey.ini', format='init')

            if self._colors is not None:
                add_to_zip(zi, self.colors, f'colors.{format}', format=format)

            if self._flux is not None:
                add_to_zip(zi, self.flux, f'flux.{format}', format=format)

            if self._light_curves is not None:
                pass

            if self._spectra is not None:
                add_to_zip(zi, self.spectra, f'spectra/', format=format)

    def __write_as_fits__(self, path):
        """
        Writes the current table into one fits file.

        :param path: The path to the save place
        :type path: str
        """
        hdu_list = [fits.PrimaryHDU(),
                    self.coordinates.to_bin_table_hdu('coordinates')]

        if self.magnitudes is not None and self.magnitudes.data is not None:
            hdu_list.extend(self.magnitudes.to_bin_table_hdu('magnitudes'))

        if self._colors is not None:
            hdu_list.extend(self.colors.to_bin_table_hdu('colors'))

        if self._flux is not None:
            hdu_list.extend(self.flux.to_bin_table_hdu('flux'))

        if self._light_curves is not None:
            hdu_list.extend(self.light_curves.to_bin_table_hdu())

        if self._spectra is not None:
            self._spectra.write(f'{os.path.dirname(path)}/spectra/')

        hdu_list = fits.HDUList(hdu_list)
        hdu_list.writeto(path, overwrite=True)

    def __write_as_csv__(self, path):
        """
        Writes the current table into one csv file.

        :param path: The path to the save place
        :type path: str
        """
        combined = self.__combine_all__()
        combined.to_csv(path)

        if self._spectra is not None:
            self._spectra.write(f'{os.path.dirname(path)}/spectra/')

    def write(self, path, format='zip', in_zip_format='fits'):
        """
        Writes the current data to a file (-system).

        For example:

        .. code:: python

            ds.write('my_data.zip')

        :param path: The path to the save place
        :type path: str
        :param format: The format of the saving (zip, fits, csv or report)
        :type format: str
        :param in_zip_format: Format of the data-files in the zip-file. Only required, if format is 'zip'
        :type in_zip_format: str
        :return:
        """
        format = format.lower()
        if format == 'zip':
            self.__write_as_zip__(path, in_zip_format)
        elif format == 'fits':
            self.__write_as_fits__(path)
        elif format == 'csv':
            self.__write_as_csv__(path)
        # elif format == 'report':
        #     self.__write_as_report__(path)
        else:
            raise ValueError(f'Format {format} is not supported.')

    def __str__(self):
        s = f'Number of entries: {len(self.coordinates)}\n'
        s += 'Available surveys:\n'
        if self._magnitudes is not None:
            if self._magnitudes.survey is not None:
                for su in self._magnitudes.survey.get_surveys():
                    s += f'\t{su}\n'
            else:
                s += '\tno survey data set\n'
        else:
            s += '\tNone\n'
        if self._colors is not None:
            s += 'Colors computed: True\n'
        else:
            s += 'Colors computed: False\n'
        if self._flux is not None:
            s += 'Flux computed: True\n'
        else:
            s += 'Flux computed: False\n'
        return s

    def __sizeof__(self):
        total_size = 56
        if self._coordinates is not None:
            total_size += sys.getsizeof(self._coordinates)
        return total_size

    @staticmethod
    def read_from_file(name, format='zip'):
        """
        Reads a DateSet from a file

        :param name: The name/path of the file
        :type name: str
        :param format: The format of the file
        :type format: str
        :return: The DataSet from the data of the file
        :rtype: DataSet
        """
        if not os.path.exists(name):
            raise FileNotFoundError(f'FileNotFoundError: [Errno 2] No such file or directory: {name}')
        if format == 'zip':
            d = {}
            head = None
            with zipfile.ZipFile(name) as zip_file:
                mags = None
                flux = FluxTable()
                colors = Colors()
                for file_name in zip_file.namelist():

                    if '.ini' not in file_name:
                        if 'magnitude' in file_name:
                            if mags is None:
                                mags = Magnitude(read_from_zip(zip_file, file_name),
                                                 survey=file_name.split('.')[0].split('magnitudes_')[-1])
                            else:
                                mags.add_survey_mags(read_from_zip(zip_file, file_name),
                                                     file_name.split('.')[0].split('magnitudes_')[-1])
                        elif 'flux' in file_name:
                            flux.add_fluxes(read_from_zip(zip_file, file_name),
                                            file_name.split('.')[0].split('magnitudes_')[-1])
                        elif 'colors' in file_name:
                            colors.add_colors(read_from_zip(zip_file, file_name),
                                              file_name.split('.')[0].split('magnitudes_')[-1])
                        elif 'coordinates' in file_name:
                            d['coordinates'] = read_from_zip(zip_file, file_name)
                        elif 'light_curves' in file_name:
                            # todo: implement light curve load
                            pass
                    else:
                        head = read_from_zip(zip_file, file_name)
                d['magnitudes'] = mags
                d['flux'] = flux
                d['colors'] = colors
            ds = DataSet(**d)
            ds._magnitudes._survey = head
            ds.reset_masks()
            return ds
        elif format == 'fits':
            with fits.open(name) as fi:
                d = {}
                for i, f in enumerate(fi):
                    if i > 0:
                        data = f.data
                        head = f.header
                        d[head['category']] = data
                return DataSet(**d)

        elif format == 'csv':
            raise NotImplementedError('Automatic csv reading not implemented')
            # todo: implement csv reading
            pass
        else:
            raise ValueError(f'Format {format} is not supported.')

    @staticmethod
    def load_coordinates(path, format='fits', ra_name='ra', dec_name='dec'):
        """
        Creates a DataSet object from a file with coordinates.

        .. code-block::

            from Phosphorpy import DataSet

            ds = DataSet.load_coordinates('./my_coordinates.fits', ra_name='RAJ2000', dec_name='DEJ2000')

        :param path: The path to the file
        :type path: str
        :param format: The format of the file
        :type format: str
        :param ra_name: The name of the RA column
        :type ra_name: str
        :param dec_name: The name of the Dec column
        :type dec_name: str
        :return: The DataSet object with the coordinates
        :rtype: DataSet
        """
        if format == 'fits':
            coords = Table.read(path)
            coords = coords[[ra_name, dec_name]]
            coords = coords.to_pandas()
            coords = coords.rename({ra_name: 'ra', dec_name: 'dec'}, axis='columns')
        elif format == 'csv':
            coords = pd.read_csv(path)[[ra_name, dec_name]]
            coords = coords.rename({ra_name: 'ra', dec_name: 'dec'}, axis='columns')
        else:
            raise ValueError(f'Format \'{format}\' is not supported.')

        return DataSet(coords)

    @staticmethod
    def from_coordinates(data=None, ra=None, dec=None):
        """
        Creates a DataSet object from a set of coordinates.
        The coordinates can be an astropy Table, astropy SkyCoord or a pandas DataFrame.
        Also ra and dec can be a list/array of coordinates.

        At least data or ra and dec must be given.

        :param data: The coordinates for the DataSet
        :type data: Table, SkyCoord, DataFrame
        :param ra: The RA-components of the coordinates or a string with the name of the RA column
        :type ra: None, ndarray, str
        :param dec: The Dec-components of the coordinates or a string with the name of the Dec column
        :type dec: None, ndarray, str
        :return: The DataSet object with the given coordinates
        :rtype: DataSet
        """
        if data is None:
            if ra is None or dec is None:
                raise ValueError('Data or ra and dec must be not None!')

            if len(ra) != len(dec):
                raise ValueError('ra and dec must have the same length.')

            data = DataFrame({'ra': ra, 'dec': dec})
        else:
            if type(data) == Table:
                data = data.to_pandas()
            elif type(data) == SkyCoord:
                data = DataFrame({'ra': data.ra.degree, 'dec': data.dec.degree})
            elif type(data) != DataFrame:
                raise NotImplementedError('Only astropy Table, astropy SkyCoord and pandas DataFrame are implement.')

            if 'ra' in data.columns and 'dec' in data.columns:
                data = data[['ra', 'dec']]
            elif ra is None or dec is None:
                raise ValueError('If the input data does not contain \'ra\' and \'dec\' ra and dec must'
                                 ' be the names of the corresponding columns.')
            else:
                data = data[[ra, dec]]
                data = data.rename({ra: 'ra', dec: 'dec'}, axis='columns')

        return DataSet(data)

    @staticmethod
    def from_vizier(name, **constrains):
        """
        Stars a constrain query on a vizier catalog and convert the output to a DataSet object.

        .. code-block::

            from Phosphorpy import DataSet

            ds = DataSet.from_vizier('SDSS', {'rmag': '<17', 'class': 'STAR'})

        :param name: The name of the survey
        :type name: str
        :param constrains:
            A dict of constrains with the key as the column name in the vizier catalog and the corresponding
            value as the actual constrain
        :return: A DataSet object with the results of the query
        :rtype: DataSet
        """
        d = constrain_query(name, **constrains)
        d = d.to_pandas()
        ds = DataSet(d)
        ds._magnitudes.add_survey_mags(d, name)
        return ds



import numpy as np
import pandas as pd
from astropy.table import Table, vstack
from astropy.coordinates import SkyCoord, Distance
from astropy import units as u

from Phosphorpy.data.sub.plots.astrometry import AstrometryPlot
from Phosphorpy.external.vizier import Gaia, BailerJones
from .table import DataTable


def _only_nearest(data):

        row_ids, row_id_count = np.unique(data['row_id'], return_counts=True)
        # find or multiple detections the closest one
        nearest = []
        for rid in row_ids[row_id_count >= 1]:
            g = data[data['row_id'] == rid]
            nearest.append(g[g['angDist'] == np.min(g['angDist'])])
        nearest = vstack(nearest)
        # data = vstack([nearest, data])
        return nearest


def _download_gaia_data(coordinates):
    """
    Downloads the Gaia data for the given coordinates

    :param coordinates: The coordinates of the targets for which Gaia data are required
    :type coordinates: Phosphorpy.data.sub.coordinates.CoordinateTable
    :return: The Gaia data
    :rtype:
    """
    g = Gaia()

    gaia = g.query(coordinates.to_astropy_table(), 'ra', 'dec', use_xmatch=True, blank=True)

    # find or multiple detections the closest one
    # gaia = _only_nearest(gaia)

    if type(gaia) == Table:
        gaia = gaia.to_pandas()
    gaia = gaia.drop_duplicates('row_id')
    gaia_coords = gaia[['row_id', 'ra', 'dec']]

    gaia = gaia.set_index('row_id')

    # join the downloaded gaia data to create empty lines if gaia doesn't provide data for a specific object
    gaia = gaia[['ra', 'ra_error', 'dec', 'dec_error',
                 'parallax', 'parallax_error',
                 'pmra', 'pmra_error',
                 'pmdec', 'pmdec_error']]
    return gaia, gaia_coords


def _download_bailer_jones_data(gaia_coords):
    """
    Downloads the distances estimated by Bailer-Jones

    :param gaia_coords: The gaia coordinates of the required targets
    :return:
    """
    # download Bailer-Jones distance estimations
    bj = BailerJones()
    if type(gaia_coords) == pd.DataFrame:
        gaia_coords = Table.from_pandas(gaia_coords)
    elif type(gaia_coords) != Table:
        gaia_coords = Table.from_pandas(gaia_coords.data)
    bj = bj.query(gaia_coords, 'ra', 'dec', use_xmatch=True, blank=True)

    bj = bj[['row_id', 'rest', 'b_rest', 'B_rest', 'rlen', 'ResFlag', 'ModFlag']]
    if type(bj) == Table:
        bj = bj.to_pandas()
    bj = bj.drop_duplicates('row_id')
    bj = bj.set_index('row_id')
    return bj


class AstrometryTable(DataTable):

    def __init__(self, mask):
        """
        AstrometryTable in the class to handle GAIA astrometry data, which includes
        coordinates, proper motion, parallax and its errors.

        :param mask: The mask object for this data
        :type mask: Phosphorpy.data.sub.table.Mask
        """
        DataTable.__init__(self, mask=mask)
        self._plot = AstrometryPlot(self)

    @staticmethod
    def load_to_dataset(ds):
        astrometry = AstrometryTable.load_astrometry(ds.coordinates)

        ds.set_astrometry(astrometry)

    @staticmethod
    def load_astrometry(coordinates):
        """
        Loads GAIA astrometry from the server.

        :param coordinates: The coordinate table with the data.
        :type coordinates: Phosphorpy.data.sub.coordinatesCoordinateTable
        :return: An AstrometryTable with the Gaia astrometry (ra, dec, parallax, pmra, pmdec and their errors)
        :rtype: AstrometryTable
        """
        gaia, gaia_coords = _download_gaia_data(coordinates)

        # download Bailer-Jones distance estimations
        bj = _download_bailer_jones_data(gaia_coords)

        # join the original gaia data with the Bailer-Jones distance data
        gaia = gaia.join(bj, how='outer')

        # df = pd.DataFrame()
        # df['row_id'] = np.arange(coordinates.data.index.values.min(),
        #                          coordinates.data.index.values.max())
        # df = df.set_index('row_id')
        # gaia = df.join(gaia, how='outer')

        astronomy = AstrometryTable(coordinates.mask)

        astronomy._data = gaia
        return astronomy

    def proper_motion(self, cos_correction=False):
        """
        Return the proper motion and the errors of it.

        :param cos_correction:
            True if the RA components should be corrected by the declination, else False.
            Default is False
        :type cos_correction: bool
        :return: pmra, pmdec, pmra_error, pmdec_error
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        temp = self._data[['pmra', 'pmdec', 'pmra_error', 'pmdec_error']].copy()
        x = temp['pmra'].values
        y = temp['pmdec'].values
        x_err = temp['pmra_error'].values
        y_err = temp['pmdec_error'].values
        if cos_correction:
            dec_rad = np.deg2rad(self._data['dec'].values)
            cos_c = np.cos(dec_rad)
            x_err = np.square(cos_c*x_err)+np.square(x*np.sin(dec_rad)*self._data['dec_error'].values/1000)
            x_err = np.sqrt(x_err)
            k = x_err > 10
            if np.sum(k) > 0:
                x_err[k] = 10
            x *= cos_c
        return pd.DataFrame({'pmra': x, 'pmdec': y, 'pmra_err': x_err, 'pmdec_err': y_err}, index=self._data.index)

    def total_proper_motion(self):
        """
        Return the total proper motion and its errors.

        .. math::
            \mu = \sqrt{\left(\mu_{\\alpha^*}\\right)^2+\mu_\delta^2}

        with

        .. math::
            \mu_{\\alpha^*} = \mu_{\\alpha}*\cos (\delta)

        :return: The total proper motion and its errors in a DataFrame
        :rtype: pandas.DataFrame
        """
        temp = self.proper_motion(True)
        pm_ra, pm_dec = temp['pmra'].values, temp['pmdec'].values
        pm_ra_error, pm_dec_error = temp['pmra_err'].values, temp['pmdec_err'].values
        pm = np.hypot(pm_ra, pm_dec)
        pm_err = np.square(pm_ra*pm_ra_error)+np.square(pm_dec*pm_dec_error)
        pm_err /= pm
        return pd.DataFrame({'pm': pm, 'err': pm_err}, index=self._data.index)

    def distance(self, kind='bailer-jones'):
        """
        Compute and return the distance in kpc

        :param kind:
            The kind of transformation. Current options are 'bailer-jones' or 'bj'
            for the distances estimated by Bailer-Jones or 'simple' for :math:`distance=1/parallax`.
            All other inputs raise a ValueError.
        :type kind: str
        :return: distance and its error in kpc
        :rtype: pd.DataFrame
        """
        kind = kind.lower()
        # if the required distance is the distance estimated by Bailer-Jones
        if kind in ('bailer-jones', 'bj'):
            lower = self.data['rest'].values-self.data['b_rest'].values
            upper = self.data['B_rest'].values-self.data['rest'].values
            return pd.DataFrame({'distance': self.data['rest'].values/1000,
                                 'error': (lower+upper)/2},
                                index=self._data.index)
        elif kind in ('simple', 'parallax'):
            distance = 1/self._data['parallax'].values
            distance_error = np.abs(1/self._data['parallax'].values**2)*self._data['parallax_error'].values
            return pd.DataFrame({'distance': distance, 'error': distance_error}, index=self._data.index)
        else:
            raise ValueError(f'Unknown kind: {kind}')

    def set_distance_limit(self, minimal, maximal, kind='bailer-jones'):
        """
        Set a limit to the distances (in kpc)

        :param minimal: The minimal distance
        :type minimal: float
        :param maximal: The maximal distance
        :type maximal: float
        :param kind:
            The kind of transformation. Current options are 'bailer-jones' or 'bj'
            for the distances estimated by Bailer-Jones or 'simple' for :math:`distance=1/parallax`.
            All other inputs raise a ValueError.
        :type kind: str
        :return:
        """
        if minimal >= maximal:
            raise ValueError(f'Minimal distance must be larger than maximal distance: {minimal} >= {maximal}.')
        d = self.distance(kind)
        d = d['distance']
        m = (d >= minimal) & (d <= maximal)
        self.mask.add_mask(m, f'Distance limit: {minimal} - {maximal}')

    def set_parallax_limit(self, minimal, maximal, with_errors=False):
        """
        Set a limit to the parallax

        :param minimal: The minimal parallax
        :type minimal: float
        :param maximal: The maximal parallax
        :type maximal: float
        :param with_errors: True if the errors should be consider too, else False. Default is False.
        :type with_errors: bool
        :return:
        """
        if minimal >= maximal:
            raise ValueError(f'Minimal value must be smaller than maximal: {minimal} > {maximal}')
        parallax = self._data['parallax']
        if with_errors:
            parallax_error = self._data['parallax_error']
            m = (parallax+parallax_error >= minimal) & (parallax-parallax_error <= maximal)
        else:
            m = (parallax >= minimal) & (parallax <= maximal)
        self.mask.add_mask(m, f'Parallax constrain: {minimal} - {maximal}')

    def set_total_proper_motion_limit(self, minimal, maximal, with_errors=False):
        """
        Set a limit to the total proper motion.
        See :meth:`total_proper_motion` fot he computing details.


        :param minimal: The minimal total proper motion in mas/yr
        :type minimal: float
        :param maximal: The maximal total proper motion in mas/yr
        :type maximal: float
        :param with_errors: True if the errors should be consider too, else False. Default is False.
        :type with_errors: bool
        :return:
        """
        if minimal >= maximal:
            raise ValueError(f'Minimal value must be smaller than maximal: {minimal} > {maximal}')

        pm = self.total_proper_motion()
        if with_errors:
            m = (pm['pm'] + pm['err'] >= minimal) & (pm['pm']-pm['err'] <= maximal)
        else:
            m = (pm['pm'] >= minimal) & (pm['pm'] <= maximal)
        self.mask.add_mask(m, f'Proper motion limit from {minimal} to {maximal}')

    def set_proper_motion_limit(self, direction, minimal, maximal, with_errors=False, cos_correction=False):
        """
        Set a limit to the proper motion

        :param direction: The direction of the proper motion (RA or Dec)
        :type direction: str
        :param minimal: The minimal proper motion
        :type minimal: float
        :param maximal: The maximal proper motion
        :type maximal: float
        :param with_errors: True if the errors should be considered in the mask, else False. Default is False.
        :type with_errors: bool
        :param cos_correction: True if the proper motion in RA direction should be declination correct, else False.
        :return:
        """
        direction = direction.lower()
        if direction != 'ra' and direction != 'dec':
            raise ValueError('Direction must be RA or Dec!')

        if minimal >= maximal:
            raise ValueError(f'Minimal value must be smaller than maximal: {minimal} > {maximal}')

        pm = self.proper_motion(cos_correction)
        pm_d = pm[f'pm{direction}']
        if with_errors:
            pm_d_err = pm[f'pm{direction}_err']
            m = (pm_d + pm_d_err >= minimal) & (pm_d-pm_d_err <= maximal)
        else:
            m = (pm_d >= minimal) & (pm_d <= maximal)
        self.mask.add_mask(m, f'Proper motion limit in {direction} direction from {minimal} to {maximal}')

    def to_sky_coord(self):
        """
        Creates from the Gaia data a astropy.coordinates.SkyCoord object for every source.
        The SkyCoord contains beside the coordinates, the proper motions (pm ra is corrected)
        and a distance estimation.

        :return: SkyCoord objects of the detections
        :rtype: SkyCoord
        """

        s = SkyCoord(self._data['ra'].values*u.deg,
                     self._data['dec'].values*u.deg,
                     pm_ra_cosdec=self._data['pmra'].values*np.cos(np.deg2rad(self._data['dec'].values))*u.mas/u.yr,
                     pm_dec=self._data['pmdec'].values*u.mas/u.yr,
                     distance=Distance(parallax=self._data['parallax'].values*u.mas))
        return s



import numpy as np
import pandas as pd
from astropy.table import Table
from astropy.io import fits
import warnings

from Phosphorpy.external import zwicky
from Phosphorpy.data.sub.plots.light_curve import LightCurvePlot
from Phosphorpy.external.css import download_light_curves


def _average_light_curve(lc, dt_max):
    """
    Averages the given light curve

    :param lc: The input light curve
    :type lc: DataFrame
    :param dt_max: The box size of the average
    :type dt_max: float
    :return: The averaged light curve
    :rtype: DataFrame
    """
    dt = lc['mjd'][1:].values - lc['mjd'][:-1].values
    p = np.where(dt > dt_max)[0]+1
    start = 0
    mags = []
    errs = []
    mjds = []
    ra = []
    dec = []
    for k in p:
        l = lc[start: k]
        err_sq = 1/l['magerr'].values**2
        err_sq_sum = 1./np.sum(err_sq)
        m = np.sum(l['mag'].values*err_sq)*err_sq_sum
        e = np.sum(err_sq*l['magerr'].values)*err_sq_sum
        mags.append(m)
        errs.append(e)
        mjds.append(np.sum(l['mjd'].values*err_sq)*err_sq_sum)
        ra.append(np.sum(l['ra'].values*err_sq)*err_sq_sum)
        dec.append(np.sum(l['dec'].values*err_sq)*err_sq_sum)
    s = lc['survey'].values[0]
    return pd.DataFrame({'mag': mags, 'magerr': errs, 'mjd': mjds,
                         'row_id': len(errs)*[lc['row_id'].values[0]],
                         'ra': ra, 'dec': dec, 'survey': np.linspace(s, s, len(errs))})


class LightCurves:
    _stat_columns = None
    _stat_operations = None

    _light_curves = None
    _average = None

    _plot = None

    def __init__(self, coordinates=None, light_curves=None, surveys=None):
        if coordinates is not None:
            if surveys is None:
                surveys = ['css', 'ptf', 'ztf']
            else:
                if type(surveys) == str:
                    surveys = [surveys]
                surveys = [s.lower() for s in surveys]

            out = []
            if 'css' in surveys:
                try:
                    css_lc = download_light_curves(coordinates['ra'],
                                                   coordinates['dec'])[['InputID', 'ID', 'Mag', 'Magerr',
                                                                        'RA', 'Decl', 'MJD']]
                    css_lc['survey'] = 1
                    css_lc = css_lc.rename({'InputID': 'row_id', 'ID': 'oid',
                                            'Mag': 'mag', 'Magerr': 'magerr', 'RA': 'ra',
                                            'Decl': 'dec', 'MJD': 'mjd'}, axis='columns')
                    out.append(css_lc)
                except:

                    warnings.warn('CSS light curve(s) are not available.')

            if type(coordinates) == pd.DataFrame:
                index = coordinates.index.values
            else:
                index = coordinates.data.index.values

            if 'ptf' in surveys:
                try:
                    ptf_lc = zwicky.download_ptf(coordinates['ra'], coordinates['dec'],
                                                 index=index)
                    ptf_lc['survey'] = 2
                    out.append(ptf_lc)
                except:
                    warnings.warn('PTF light curve(s) are not available.')

            if 'ztf' in surveys:
                try:
                    zwicky_lc = zwicky.download_light_curve(coordinates['ra'], coordinates['dec'],
                                                            index=index)
                    zwicky_lc['survey'] = 3
                    out.append(zwicky_lc)
                except:
                    warnings.warn('ZTF light curve(s) are not available.')

            if len(out) == 0:
                raise ValueError('No light curves found for the coordinate and survey combination.')

            self._light_curves = pd.concat(out)
        elif light_curves is not None:
            self._light_curves = light_curves
        else:
            raise ValueError('coordinates or light curves must be given.')

        self._stat_columns = ['row_id', 'survey', 'mag', 'magerr', 'ra', 'dec', 'mjd']
        self._stat_operations = [np.mean, np.median, np.std, np.min, np.max, 'count']

        self._plot = LightCurvePlot(self)

    def __str__(self):
        out = ''.join(['Number of light curves: {}\n',
                       'with {} entries.'])
        out = out.format(len(np.unique(self._light_curves['row_id'])), len(self._light_curves))
        return out

    def __repr__(self):
        return self.__str__()

    def stats(self):
        """
        Computes the statistics of the light curves

        :return: The statistics of the light curves
        :rtype: pandas.DataFrame
        """
        return self.light_curves[self._stat_columns].groupby(['row_id', 'survey']).aggregate(self._stat_operations)

    def average(self, dt_max=1, overwrite=False):
        """
        Averages the light curve

        :param dt_max: The maximal time difference between two data points
        :type dt_max: float
        :param overwrite: True, if the previous results should be overwritten, else False. Default is False.
        :type overwrite: bool
        :return: The averaged light curves
        :rtype: LightCurves
        """
        dt_min = np.min(self._light_curves['mjd'].values[1:]-self._light_curves['mjd'].values[:-1])
        if dt_max < 0:
            raise ValueError('dt_max must be larger than 0')
        elif dt_max < dt_min:
            raise ValueError(f'dt_max must be larger than the minimal difference between two data-points'
                             f'({dt_min})')

        if self._average is not None and overwrite:
            return self._average

        out = []
        for lc_id in np.unique(self._light_curves['row_id']):
            lc = self._light_curves[self._light_curves['row_id'] == lc_id]
            for s in np.unique(lc['survey']):

                out.append(_average_light_curve(lc[lc['survey'] == s], dt_max))
        self._average = LightCurves(light_curves=pd.concat(out))
        return self._average

    def get_light_curve(self, index):
        """
        Returns the data of the light curve with the given index

        :param index: The index of the light curve
        :rtype index: int
        :return:
        """
        return LightCurves(light_curves=self._light_curves[self._light_curves['row_id'] == index])

    def align_light_curves(self, inplace=True):
        """
        Aligns the median of the light curves from the different surveys to median of the CSS light curve

        .. warning::

            Because of the simple median alignment, the aligned light curves can have systematic effects.
            For example, if a light curve as a general trend over the time of one survey, the correction will
            fail. This can happen, if light curves of QSO's are used or if a transient is present in one of the
            light curves without enough coverage before and after it.

        :param inplace: True, if the stored light curve should be overwritten, else False. Default is True.
        :type inplace: bool
        :return: The aligned light curves, if inplace is True. Otherwise None.
        :rtype: LightCurves, None
        """
        lc = self._light_curves.copy()
        stats = self.stats()
        mag0 = 0
        out = []
        for index in stats.index.values:
            mag_med = stats.loc[index][('mag', 'median')]
            lc_mask = (lc['row_id'] == index[0]) & (lc['survey'] == index[1])
            if index[1] == 1:
                mag0 = mag_med
                out.append(lc[lc_mask])
            else:
                lc_s = lc[lc_mask]
                lc_s.loc[:, 'mag'] -= mag_med-mag0
                out.append(lc_s)
        out = pd.concat(out)

        if inplace:
            self._light_curves = out
        else:
            return LightCurves(light_curves=out)

    def to_time_series(self, index):
        raise NotImplementedError()

    def to_astropy_table(self):
        """
        Returns the light curve data as an astropy Table

        :return: The light curve data points
        :rtype: Table
        """
        return Table.from_pandas(self.light_curves)

    def to_bin_table_hdu(self):
        return fits.BinTableHDU(self.to_astropy_table())

    @property
    def light_curves(self):
        return self._light_curves

    @property
    def data(self):
        return self.light_curves

    @property
    def stat_columns(self):
        return self._stat_columns

    @stat_columns.setter
    def stat_columns(self, value):
        if type(value) == str:
            if value in self.light_curves.columns:
                self._stat_columns = value
            else:
                raise ValueError(f'{value} is not a column.')
        elif type(value) == (list or tuple):
            for v in value:
                if v not in self.light_curves.columns:
                    raise ValueError(f'{v} is not a columns.')
        else:
            raise ValueError('Str or tuple/list are allowed.')

    @property
    def plot(self):
        return self._plot

    def write(self, path, format='fits'):
        """
        Writes the light curve data to a file

        :param path: The path to the file
        :type path: str
        :param format: The format of the file ('fits' or 'csv')
        :type format: str
        :return:
        """
        if format == 'fits':
            Table.from_pandas(self.light_curves).write(path, overwrite=True)
        elif format == 'csv':
            self.light_curves.to_csv(path)
        else:
            raise ValueError(f'{format} is not supported.')

    @staticmethod
    def read(path, format='fits'):
        """
        Reads the light curve data from a fits or csv file

        :param path: The path to the file
        :type path: str
        :param format: The format of the file (fits or csv)
        :type format: str
        :return: The light curve data from the file
        :rtype: LightCurves
        """
        if format == 'fits':
            return LightCurves(light_curves=Table.read(path).to_pandas())
        elif format == 'csv':
            return LightCurves(light_curves=pd.read_csv(path, index_col=0))
        else:
            raise ValueError(f'{format} is not supported.')



import numba as nb
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.units.quantity import Quantity
from pandas import DataFrame
from sklearn.neighbors import NearestNeighbors

from .plots.coordinates import CoordinatePlot
from .table import DataTable

RA_NAMES = ['ra', 'Ra', 'RA', 'RAJ2000', 'RA_ICRS', '_RAJ2000']
DEC_NAMES = ['dec', 'Dec', 'DEC', 'DEJ2000', 'DECJ2000', 'DE_ICRS', '_DEJ2000']
L_NAMES = ['l', 'L']
B_NAMES = ['b', 'B']


class CoordinateTable(DataTable):
    __slots__ = ('_data', '_head', '_plot')

    def __init__(self, data, head=None, mask=None):
        """
        Table to handle coordinates

        :param data:
        :param head:
        """
        DataTable.__init__(self, mask=mask)
        self._head = head
        self._plot = CoordinatePlot(self)

        # create SkyCoord objects to get the galactic coordinates of the sources
        if type(data) == SkyCoord:
            ra = np.array(data.ra.degree)
            dec = np.array(data.dec.degree)
            lon = np.array(data.galactic.l)
            b = np.array(data.galactic.b)
        elif type(data) == DataFrame:
            s = SkyCoord(
                data['ra'].values*u.deg,
                data['dec'].values*u.deg
            )
            ra = s.ra.degree
            dec = s.dec.degree
            lon = s.galactic.l.degree
            b = s.galactic.b.degree
        else:
            if type(data) != np.ndarray or len(data.shape) == 1:
                columns = get_column_names(data)

                s = SkyCoord(np.array(data[find_ra_column(columns)]) * u.deg,
                             np.array(data[find_dec_column(columns)]) * u.deg)
            else:
                s = SkyCoord(data[:, 0] * u.deg, data[:, 1] * u.deg)
            ra = np.array(s.ra.degree)
            dec = np.array(s.dec.degree)
            lon = np.array(s.galactic.l)
            b = np.array(s.galactic.b)
        d = np.zeros((len(data), 4))
        d[:, 0] = ra
        d[:, 1] = dec
        d[:, 2] = lon
        d[:, 3] = b
        self._data = DataFrame(data=d, columns=['ra', 'dec', 'l', 'b'],
                               index=np.arange(len(ra))+1)
        self._data['row_id'] = np.arange(len(self._data))
        self._data = self._data.set_index('row_id')

    def __getitem__(self, item):
        if type(item) == str:
            # check if the item is one of the default ones
            if item in RA_NAMES:
                return self._data['ra'].values
            elif item in DEC_NAMES:
                return self._data['dec'].values
            elif item in L_NAMES:
                return self._data['l'].values
            elif item in B_NAMES:
                return self._data['b'].values
            else:
                raise KeyError('Key {} not known! Choose one of the default ones.'.format(item))
        elif type(item) == int:
            return self._data.loc[item]

    def __eq__(self, other):
        if type(other) == SkyCoord or type(other) == np.ndarray or type(other) == CoordinateTable:
            if len(self) == len(other):
                # of the matching doesn't include a -1, which means that there is no match,
                # the coordinate sets are equal
                if len(self.match(other)) == len(self):
                    return True
                else:
                    return False
            else:
                return False
        else:
            raise ValueError('Allowed types are astropy.coordinate.SkyCoord, numpy.ndarray '
                             'or SearchEngine.data.sub.coordinates.CoordinateTable, not {}'.format(type(other)))

    def __len__(self):
        return len(self._data)

    def to_table(self, galactic=False):
        """
        Returns the coordinates as pandas.DataFrame

        :param galactic: True if the galactic coordinates should be included, else False. Default is False.
        :type galactic: bool
        :return: The coordinates in a DataFrame
        :rtype: pandas.DataFrame
        """
        if galactic:
            return self.data
        else:
            return self.data[['ra', 'dec']]

    def to_sky_coord(self, source_id=-1):
        """
        Return the coordinate(s) back as a SkyCoord object

        :param source_id:
            The id of the source. Default is -1, which means that all coordinates are converted to SkyCoord.
        :type source_id: int, list, tuple
        :return: The SkyCoord object of the source(s)
        :rtype: astropy.coordinates.SkyCoord
        """
        if source_id == -1:
            s = SkyCoord(self.data['ra'].values * u.deg, self.data['dec'].values * u.deg)
        else:
            s = SkyCoord(self.data['ra'].values[source_id] * u.deg,
                         self.data['dec'].values[source_id] * u.deg)
        return s
    
    def to_astropy_table(self, category='coordinates', **kwargs):
        return super(CoordinateTable, self).to_astropy_table(category, **kwargs)

    def _combine_coordinates(self, other):

        # prepare the input coordinates for the match
        if type(other) == SkyCoord:
            d = np.zeros((len(other), 2))
            d[:, 0] = other.ra.degree
            d[:, 1] = other.dec.degree
        elif type(other) == CoordinateTable or type(other) == Table or type(other) == DataFrame:
            d = np.zeros((len(other), 2))
            d[:, 0] = other['ra']
            d[:, 1] = other['dec']
        elif type(other) == np.ndarray:
            d = other.copy()
        else:
            raise ValueError()

        x = np.concatenate((self._data[['ra', 'dec']].values, d), axis=0)
        return x

    def match(self, other, radius=2 * u.arcsec,
              ra_name='ra', dec_name='dec'):
        """
        Matches these coordinates with another coordinate set. To do that it uses a nearest neighbor algorithm
        to find the next neighbor.
        The distance is computed with the approximation of small angles

        .. math::

            d = \sqrt{\left(\Delta\\alpha \cdot \cos\left(\delta\\right)\\right)^2 + \left(\Delta \delta\\right)^2}

        as an additional approximation because performance reasons, in this algorithm its assumed that

        .. math::

            \cos \delta_1 \\approx \cos \delta_2

        and therefore we can use

        .. math::

            \Delta \\alpha \cdot \cos \delta = \\alpha_1 \cdot \cos \delta - \\alpha_2 \cdot \cos \delta \\\\
            \\approx \\alpha_1 \cos\delta_1 - \\alpha_2 \cos \delta_2

        :param other: The second coordinate set to match with it
        :type other: numpy.ndarray, astropy.coordinates.SkyCoord, Phosphorpy.data.sub.coordinates.CoordinateTable
        :param radius: The cross-match radius as float or astropy.unit. If it is a float, it will be taken as degree.
        :type radius: float, astropy.units
        :param ra_name: Name of the RA column. Default is 'ra'.
        :type ra_name: str
        :param dec_name: Name of the Dec column. Default is 'dec'.
        :type dec_name: str
        :returns: An array with the indices of the match sources. Sources without a match will have -1.
        :rtype: pandas.DataFrame
        """
        x = self._data[['ra', 'dec']].values
        # convert the radius to degree and to a float
        if type(radius) == Quantity:
            radius = radius.to(u.deg)

        if type(other) != DataFrame:
            if type(other) == Table:
                other = other.to_pandas()
            elif type(other) == np.ndarray:
                other = DataFrame(other, columns=['ra', 'dec'])
            elif type(other) == CoordinateTable:
                other = other.data
            elif type(other) == SkyCoord:
                sk = self.to_sky_coord()
                index, dis2d, dis3d = sk.match_to_catalog_sky(other)
                return other[index[dis2d < radius]]
            else:
                raise ValueError('Unsupported input type ({})'.format(str(type(other))))

        # approximate small distance
        x[:, 0] *= np.cos(np.deg2rad(x[:, 1]))
        nn = NearestNeighbors(n_neighbors=1).fit(x)

        y = other[[ra_name, dec_name]].values
        y[:, 0] *= np.cos(np.deg2rad(y[:, 1]))
        distances, indices = nn.kneighbors(y)
        m = distances[:, 0] <= radius.value

        new_indices = self.data.index.values[indices[:, 0][m]]

        # select the matched sources only and sort them in the same order
        other = other[m]
        other['row_id'] = new_indices
        other = other.set_index('row_id')
        return other

    def get_closest_source_id(self, coordinate=None, ra=None, dec=None):
        """
        Returns the ID of the closest sources and the distance to source.

        coordinate or ra and dec has to be given. If all three are given,
        the coordinate object is taken.

        :param coordinate: Coordinates (or close to) of the required ID.
        :type coordinate: SkyCoord
        :param ra: The RA component of the coordinates
        :type ra: float
        :param dec: The Dec component of the coordinates
        :type dec: float
        :return: The closest ID and the distance to the corresponding source
        :rtype: int, u.Quantity
        """
        if coordinate is None:
            if ra is None or dec is None:
                raise ValueError('If no coordinates are given, RA and Dec must be given.')
            coordinate = SkyCoord(ra*u.deg, dec*u.deg)
        distances = coordinate.separation(self.to_sky_coord())
        min_id = np.argmin(distances)
        return min_id, distances[min_id]

    def write(self, path, data_format='parquet'):
        data = self.to_table()
        if data_format == 'parquet':
            data.to_parquet(path)
        elif data_format == 'csv':
            data.to_csv(path)
        elif data_format == 'latex':
            data.to_latex(path)
        elif data_format == 'fits':
            Table.from_pandas(data).write(path, overwrite=True)


def get_column_names(d):
    """
    Returns the name of the columns
    :param d: The input data set
    :type d: numpy.ndarray, astropy.table.Table, pandas.DataFrame
    :return: A list with the column names
    :rtype: list
    """
    if type(d) == np.ndarray:
        return d.dtype.names
    elif type(d) == Table:
        return d.colnames
    elif type(d) == DataFrame:
        return d.columns


def find_ra_column(columns):
    """
    Search in the columns for a possible RA name

    :param columns: The column names of the input data set
    :type columns: list
    :return: The name of the RA column, if it is in. If the name is not included it will raise a KeyError
    :rtype: str
    """
    for r in RA_NAMES:
        if r in columns:
            return r
    raise KeyError('RA column not found!')


def find_dec_column(columns):
    """
    Search in the columns for a possible Dec name

    :param columns: The column names of the input data set
    :type columns: list
    :return: The name of the Dec column, if it is in. If the name is not included it will raise a KeyError
    :rtype: str
    """
    for d in DEC_NAMES:
        if d in columns:
            return d
    raise KeyError('Dec column not found!')


@nb.jit('f8(f8[:], f8[:])', nopython=True)
def spherical_distance(x1, x2):
    """
    Computes the spherical distance in the approximation of small distances

    .. math::

        d = \sqrt{x1\cdot \cos\left(x2\\right) + x2}

    :param x1: The first vector
    :type x1: Union
    :param x2: The second vector
    :type x2: Union
    :return: THe distance between x1 and x2
    :rtype: float
    """
    x = x1 - x2
    return np.sqrt((x[0] * np.cos(np.deg2rad(x1[1]))) ** 2 + (x[1]) ** 2)



import numpy as np
import pylab as pl
import warnings

def _hist(sp, x, bins, histtype, label=''):

    sp.hist(x, bins=bins, histtype=histtype, label=label, range=[np.nanmin(x),
                                                                 np.nanmax(x)])


class MagnitudePlot:
    _data = None

    def __init__(self, data):
        self._data = data

    def hist(self, cols=None, survey=None, path=''):
        """
        Plots the histogram(s) of the different magnitude(s).

        :param cols: A list of magnitude names. Default is None which means that all magnitudes are taken.
        :type cols: list, str
        :param survey: Name of the survey or None. If None all surveys with the cols names are used.
        :type survey: str
        :param path: Path to the save place. Default is an empty string, which means that the plot will be shown, only.
        :type path: str
        :return:
        """
        if survey is not None:
            d = self._data.get_survey_data(survey)
        elif len(self._data.data) == 1:
            d = self._data.data[0]
        else:
            d = self._data.get_columns(cols)
        all_columns = d.columns.values
        # if no columns are given, take all columns from the dataset
        if cols is None:
            cols = all_columns
        elif type(cols) == str:
            cols = [cols]

        # check every input column if it contains 'mag' in its name, if not add 'mag'
        # to the end of the column name to convert it to the default style
        for i, c in enumerate(cols):
            if 'mag' not in c:
                cols[i] = f'{c}mag'

        pl.clf()
        sp = pl.subplot()

        column_not_found = 0
        # go through all given magnitudes
        for c in cols:
            # check if the magnitude is in the data. If not, raise an error
            if c not in all_columns:
                warnings.warn(f'Magnitude ({c}) is not in the Dataset!')
                column_not_found += 1
                continue

            if self._data.mask.get_mask_count() == 0:
                _hist(sp, d[c],
                      bins='auto', histtype='step',
                      label=c.split('mag')[0])
            else:
                for i in range(self._data.mask.get_mask_count()):
                    _hist(sp, d[c][self._data.mask.get_mask(i).values],
                          bins='auto', histtype='step', label=c)

        if column_not_found == len(cols):
            raise ValueError('At least oe column must be in one survey available.')
        
        # set the axis-labels
        sp.set_xlabel('magnitude')
        sp.set_ylabel('counts')
        pl.legend(loc='best')

        # if a path is given, save the figure, else show the figure
        if path != '':
            pl.savefig(path)
        else:
            pl.show()



import numpy as np
import numba


def power_2_10(x):
    return np.power(10., -x/2.5)


@numba.vectorize
def subtract(a, b):
    """
    Numba implementation of a subtraction

    :param a:
    :param b:
    :return:
    """
    return a-b


def gaus(x, a, b, c, d, e):
    """
    Fitting function with a gaussian component, a linear component and a static component

    :math:`f(x) = a*\exp\left(-\\frac{(x-x_0)^2}{2*\sigma^2}\\right)+b*x+c`

    :param x: The variable values
    :param a: The strength of the gaussian
    :param b: The shift of the center
    :param c: The sigma of the gaussian
    :param d: The steepness rate of the linear component
    :param e: The value of the static component
    :return: The corresponding function values
    :rtype: np.ndarray
    """
    return a * np.exp(-np.square(x-b)/(2*c**2))+e+d*x


def smooth2d(mat, c=5):
    """
    Smooths a 2d-array

    :param mat: The input data
    :type mat: np.ndarray
    :param c: The number of smooths
    :param c: int
    :return: The c-times smoothed input data
    """
    if c == 0:
        return mat
    out = np.zeros(mat.shape)

    out[0, 0] = (2*mat[0, 0]+mat[1, 0]+mat[0, 1])/4
    out[0, -1] = (2*mat[0, -1]+mat[1, -1]+mat[0, -2])/4
    out[-1, 0] = (2*mat[-1, 0]+mat[-2, 0]+mat[-1, 1])/4
    out[-1, -1] = (2*mat[-1, -1]+mat[-2, -1]+mat[-1, -2])/4

    out[1:-1, 0] = (3*mat[1:-1, 0]+mat[1:-1, 1]+mat[:-2, 0]+mat[2:, 0])/6
    out[1:-1, -1] = (3*mat[1:-1, -1]+mat[1:-1, -2]+mat[:-2, -1]+mat[2:, -1])/6
    out[0, 1:-1] = (3*mat[0, 1:-1]+mat[1, 1:-1]+mat[0, :-2]+mat[0, 2:])/6
    out[-1, 1:-1] = (3*mat[-1, 1:-1]+mat[-2, 1:-1]+mat[-1, :-2]+mat[-1, 2:])/6

    out[1:-1, 1:-1] = (4*mat[1:-1, 1:-1]+mat[:-2, 1:-1]+mat[2:, 1:-1]+mat[1:-1, :-2]+mat[1:-1, 2:])/8
    return smooth2d(out, c=c-1)



import numpy as np
import pylab as pl


class AstrometryPlot:
    _astrometry = None

    def __init__(self, astrometry):
        self._astrometry = astrometry

    def proper_motion(self, path='', cos_correction=False):
        """
        Makes a plot of the proper motion

        :param path: Path to the location
        :type path: str
        :param cos_correction: True if the pmra should be correct with cos(deg), else False. Default is False.
        :type cos_correction: bool
        :return:
        """
        pms = self._astrometry.proper_motion(cos_correction)

        pl.clf()
        sp = pl.subplot()
        sp.errorbar(pms['pmra'], pms['pmdec'], xerr=pms['pmra_err'], yerr=pms['pmdec_err'], fmt='.', capsize=2)
        for i in range(1, self._astrometry.mask.get_mask_count()):
            m = self._astrometry.mask.get_mask(i).copy()
            m.index.name = pms.index.name
            dpms = pms[m]
            sp.errorbar(dpms['pmra'], dpms['pmdec'], xerr=dpms['pmra_err'], yerr=dpms['pmdec_err'], fmt='.', capsize=2)

        if cos_correction:
            sp.set_xlabel('$\\mu_\\alpha^*$ [mas/yr]')
        else:
            sp.set_xlabel('$\\mu_\\alpha$ [mas/yr]')
        sp.set_ylabel('$\\mu_\\delta$ [mas/yr]')

        if path != '':
            pl.savefig(path)
        else:
            pl.show()

    def _hist(self, x, xlabel, path=''):
        """
        Makes a histogram plot of the input data

        :param x: The input data
        :type x: Union, numpy.ndarray
        :param xlabel: The x-label
        :type xlabel: str
        :param path: Path to the location
        :type path: str
        :return:
        """
        pl.clf()
        sp = pl.subplot()
        hist_range = [np.nanmin(x), np.nanmax(x)]
        sp.hist(x, bins='auto', range=hist_range, histtype='step', color='k')
        for i in range(1, self._astrometry.mask.get_mask_count()):
            sp.hist(x[self._astrometry.mask.get_mask(i)].values,
                    bins='auto', range=hist_range, histtype='step')
        sp.set_xlabel(xlabel)
        sp.set_ylabel('count')

        if path != '':
            pl.savefig(path)
        else:
            pl.show()

    def proper_motion_hist(self, path=''):
        """
        Plots a histogram of the total proper motion

        :param path: The path to the location
        :type path: str
        :return:
        """
        pms = self._astrometry.proper_motion(True)
        pm = np.hypot(pms['pmra'], pms['pmdec'])
        self._hist(pm, '$\\mu$ [mas/yr]', path)

    def parallax_hist(self, path=''):
        """
        Plots a histogram of the parallax

        :param path: The path to the location
        :type path: str
        :return:
        """
        parallax = self._astrometry.data['parallax']
        self._hist(parallax, 'p [mas]', path)

    def parallax(self, path=''):
        """
        Plot the parallax vs its error

        :param path: The path to the location
        :type path: str
        :return:
        """
        pl.clf()
        sp = pl.subplot()
        parallax = self._astrometry.data['parallax']
        parallax_error = self._astrometry.data['parallax_error']
        sp.scatter(parallax, parallax_error, marker='.', c='k')

        for i in range(1, self._astrometry.mask.get_mask_count()):
            m = self._astrometry.mask.get_mask(i)
            sp.scatter(parallax[m].values, parallax_error[m].values, marker='.')

        sp.set_xlabel('p [mas]')
        sp.set_ylabel('$\\sigma_p$ [mas]')

        if path != '':
            pl.savefig(path)
        else:
            pl.show()
